#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" User Interface Options model module."""
from cache_memoize import cache_memoize

from django.db import models
from django.utils.translation import gettext_lazy as _

from .base_model import BaseModel
from .institution import Institution
from .user import InstitutionUser
from .user import get_institution_user
from .user import get_institution_roles
from .user import is_global_admin


@cache_memoize(24 * 60 * 60)
def get_role_base_ui_routes():
    """
        Compute the available routes for each role
        :return: List of allowed routes per role
        :rtype: dict
    """
    base_routes = UIOption.objects.filter(enabled=True,
                                          institution=None,
                                          roles=None
                                          ).values_list('route', flat=True)
    # Group each route by role
    routes = {
        '__all__': set(base_routes)
    }
    role_required_routes = UIOption.objects.filter(enabled=True,
                                                   institution=None,
                                                   roles__isnull=False
                                                   ).values_list('route', 'roles')
    for route, roles in role_required_routes:
        role_list = roles.split(',')
        for route_role in role_list:
            if route_role not in routes:
                routes[route_role] = set()
            routes[route_role].add(route)

    # Convert sets to lists to allow serialization
    for role in routes:
        routes[role] = list(routes[role])

    return routes


def get_user_ui_routes(user):
    """
        Compute the user interface options for a user

        :param user_id: The user id
        :type user_id: int
        :return: List of allowed routes
        :rtype: list
    """
    inst_user = get_institution_user(user)

    # Users not belonging to any institutions can see nothing unless are global_admins
    if inst_user is None and not is_global_admin(user):
        return []

    role_routes = get_role_base_ui_routes()
    routes = set(role_routes['__all__'])

    # Add routes for Global Admins
    if is_global_admin(user) and 'GLOBAL_ADMIN' in role_routes:
        routes = routes.union(set(role_routes['GLOBAL_ADMIN']))

    # If user has no institution, we can finish
    if inst_user is None:
        return list(routes)

    # Add routes for other roles
    for role in get_institution_roles(user):
        if role in role_routes:
            routes = routes.union(set(role_routes[role]))

    inst_blocked_routes = UIOption.objects.filter(institution=inst_user.institution,
                                                  enabled=False).values_list('route', flat=True)

    user_blocked_routes = UIOption.objects.filter(institution=inst_user.institution,
                                                  user=inst_user,
                                                  enabled=False).values_list('route', flat=True)

    allowed_routes = routes - set(inst_blocked_routes) - set(user_blocked_routes)

    return list(allowed_routes)


class UIOption(BaseModel):
    """ UI Option model. """
    route = models.CharField(max_length=250, null=False, blank=False, db_index=True,
                             help_text=_("Affected Route."))

    enabled = models.BooleanField(null=None, blank=None, default=False,
                                  help_text=_("Status"))

    institution = models.ForeignKey(Institution, null=True, on_delete=models.CASCADE, default=None,
                                    help_text=_('Affected institution'))

    user = models.ForeignKey(InstitutionUser, null=True, on_delete=models.CASCADE, default=None,
                             help_text=_('Affected user'))

    roles = models.CharField(max_length=250, null=True, blank=True,
                             help_text=_("Required roles."))

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def __repr__(self):
        return "<UIOption(id='%r', route='%r', enabled='%r', user='%r', institution='%r')>" % (
            self.id, self.route, self.enabled, self.user, self.institution)

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        super().save(force_insert, force_update, using, update_fields)
        # Invalidate cached value
        if self.institution is None:
            get_role_base_ui_routes.invalidate()



# Generated by Django 3.0 on 2021-05-06 20:30

import django.db.models.deletion
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('tesla_ce', '0012_auto_20210415_1829'),
    ]

    operations = [
        migrations.CreateModel(
            name='WebhookClient',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='Name that identifies the client.', max_length=250, unique=True)),
                ('client_header', models.CharField(help_text='Name of the request header used to identify the client.', max_length=250, unique=True)),
                ('id_header', models.CharField(default=None, help_text='Name of the request header used to get the message id.', max_length=250, null=True)),
                ('description', models.TextField(blank=True, help_text='Client description.', null=True)),
                ('implementation', models.CharField(blank=True, help_text='Webhook implementation class.', max_length=255)),
                ('enabled', models.BooleanField(blank=None, default=False, help_text='Whether this client is enabled or not', null=None)),
                ('credentials', models.TextField(help_text='Client authentication credentials.')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='WebhookMessage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', models.IntegerField(choices=[(0, 'Created'), (1, 'Processing'), (2, 'Processed'), (3, 'Error'), (4, 'Timeout')], default=0, help_text='Status for this request')),
                ('message_id', models.CharField(help_text='Message Id.', max_length=250)),
                ('error', models.TextField(blank=True, help_text='Error message.', null=True)),
                ('body', models.TextField(help_text='Webhook message body.')),
                ('start_process_at', models.DateTimeField(default=None, help_text='When the webhook started to be processed', null=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('client', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.WebhookClient')),
            ],
            options={
                'unique_together': {('client', 'message_id')},
            },
        ),
    ]



#  Copyright (c) 2021 Mireia Bellot
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Test module for institution Instructor data management """
import logging

import pytest

import tests.utils


@pytest.mark.django_db
def test_api_institution_instructor(rest_api_client, institution_course_test_case):
    # Institution users need mail domain validation
    institution_course_test_case['institution'].mail_domain = 'tesla-ce.eu'
    institution_course_test_case['institution'].save()
    institution_user = institution_course_test_case['user'].institutionuser
    institution_id = institution_course_test_case['institution'].id

    # List Institution Instructor List
    logging.info('\n1) List Institution Instructor List --------------------------------------')
    # Set Institution Admin privileges
    institution_user.inst_admin = True
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)

    str_path = '/api/v2/institution/{}/instructor/'.format(institution_id)
    str_message = 'RESPONSE institution_id:{}'.format(institution_id)
    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'List Institution Instructor List ', str_message, 200)
    n_instructors = body['count']

    # Add a new Instructor to an institution
    logging.info('\n2) Add a new Instructor to an institution --------------------------------------')

    # 666 IntegrityError(1048, "Column 'last_name' cannot be null")
    str_data = {'uid': 'INSTRUCTOR_UID_TEST', 'first_name': 'INSTRUCTOR FIRST NAME',
                'last_name': 'INSTRUCTOR LAST NAME', 'email': 'email@tesla-ce.eu'}
    new_instructor_id = tests.utils.post_rest_api_client(rest_api_client, str_path, str_data,
                                                         'Add a new Instructor to an institution',
                                                         'RESPONSE: ', 201)

    #  Status 400: Validating email domain
    str_data = {'uid': 'INSTRUCTOR_UID_TEST', 'first_name': 'INSTRUCTOR FIRST NAME',
                'last_name': 'INSTRUCTOR LAST NAME', 'email': 'email@email.com'}
    tests.utils.post_rest_api_client(rest_api_client, str_path, str_data,
                                     'Add a new Instructor to an institution',
                                     'RESPONSE: ', 400)

    # Ensure number of instructors has increased
    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'List Institution Instructor List ', str_message, 200)
    assert n_instructors + 1 == body['count']

    # Read Instructor information
    logging.info('\n3) Read Instructor information --------------------------------------')
    str_path = '/api/v2/institution/{}/instructor/{}/'.format(institution_id, new_instructor_id)
    str_response = 'RESPONSE Instructor ID={}:'.format(new_instructor_id)
    tests.utils.get_rest_api_client(rest_api_client, str_path,
                                    'Read Instructor information', str_response, 200)

    # Update Instructor information
    logging.info('\n4) Update Instructor information --------------------------------------')

    str_data = {'uid': 'INSTRUCTOR_UID_TEST', 'first_name': 'CHANGED_INSTRUCTOR FIRST NAME',
                'last_name': 'CHANGED_INSTRUCTOR LAST NAME', 'email': 'CHANGED_email@tesla-ce.eu'}
    tests.utils.put_rest_api_client(rest_api_client, str_path, str_data,
                                    'Update Instructor Information', str_response, 200)

    # Delete Instructor from an Institution
    logging.info('\n5) Delete Instructor from an Institution --------------------------------------')

    str_path = '/api/v2/institution/{}/instructor/{}/'.format(institution_id, new_instructor_id)
    tests.utils.delete_rest_api_client(rest_api_client, str_path,
                                       'Delete Instructor from an Institution', "RESPONSE: ", 204)

    # Ensure number of Instructors has decreased
    str_path = '/api/v2/institution/{}/instructor/'.format(institution_id)
    str_message = 'RESPONSE institution_id:{}'.format(institution_id)
    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'List Institution Instructor List', str_message, 200)
    assert body['count'] == n_instructors

    # Ensure Instructors cannot be listed with normal user
    # Remove Institution Admin privileges
    institution_user.inst_admin = False
    institution_user.send_admin = False
    institution_user.legal_admin = False
    institution_user.data_admin = False
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)

    logging.info('\n6) Listing Institution Instructor List with invalid user  --------------------------------------')
    str_path = '/api/v2/institution/{}/instructor/'.format(institution_id)
    str_message = 'RESPONSE institution_id:{}'.format(institution_id)
    tests.utils.get_rest_api_client(rest_api_client, str_path,
                                    'List Institution Instructor List with invalid user', str_message, 403)

    # Add an Instructor to an Institution with invalid user
    logging.info('\n7) Add an Instructor to an Institution with invalid user  --------------------------------------')
    str_data = {'username': 'USERNAME_TEST', 'email': 'email@email.com'}
    tests.utils.post_rest_api_client(rest_api_client, str_path, str_data,
                                     'Add an Instructor to an Institution with invalid user', 'RESPONSE: ', 403)



# Generated by Django 3.0.6 on 2020-05-20 19:28

import django.db.models.deletion
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('tesla_ce', '0005_auto_20200520_2009'),
    ]

    operations = [
        migrations.AlterField(
            model_name='assessmentsessiondata',
            name='session',
            field=models.OneToOneField(help_text='Related assessment session', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.AssessmentSession'),
        ),
    ]



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Assessment session model module."""
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from .activity import Activity
from .base_model import BaseModel
from .learner import Learner


class AssessmentSession(BaseModel):
    """ Assessment session model. """
    activity = models.ForeignKey(Activity, null=False,
                                 on_delete=models.CASCADE, help_text=_('Activity of this session'))

    learner = models.ForeignKey(Learner, null=False,
                                on_delete=models.CASCADE, help_text=_('Learner of this session'))

    started_at = models.DateTimeField(auto_now_add=True,
                                      help_text=_('Date when this assessment session started'))
    checked_at = models.DateTimeField(null=True, blank=True,
                                      help_text=_('Last time this assessment session has been checked'))
    closed_at = models.DateTimeField(null=True, blank=True,
                                     help_text=_('Date the assessment session has been closed'))
    auto_closed = models.BooleanField(null=True, default=None,
                                      help_text=_('Whether this session has been automatically closed'))

    @property
    def is_active(self):
        """
            Whether this assessment session is active or not

            :return: True if it is active or False if it is closed
            :rtype: bool
        """
        return self.closed_at is None

    def close(self, auto=False, close_related=True):
        """
            Close an activity session

            :param auto: Whether this close action has been automatically triggered
            :param close_related: Close open sessions for same learner and activity
        """
        # Close this session
        self.closed_at = timezone.now()
        self.auto_closed = auto
        self.save()

        # Close related sessions
        if close_related:
            for session in AssessmentSession.objects.filter(activity=self.activity,
                                                            learner=self.learner,
                                                            started_at__lt=self.started_at,
                                                            closed_at=None):
                session.close(auto=True, close_related=False)

    def __repr__(self):
        return "<AssessmentSession(activity_id='%r',learner_id='%r', is_active='%r')>" % (
            self.activity_id, self.learner_id, self.is_active)



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
from .activity import InstitutionCourseActivitySerializer
from .activity import InstitutionCourseActivityExtendedSerializer
from .activity_instrument import InstitutionCourseActivityInstrumentSerializer
from .activity_report import InstitutionCourseActivityReportDetailSerializer
from .activity_report import InstitutionCourseActivityReportExtendedSerializer
from .activity_report import InstitutionCourseActivityReportLearnerSerializer
from .activity_report import InstitutionCourseActivityReportSerializer
from .activity_report_audit import InstitutionCourseActivityReportAuditSerializer
from .course import InstitutionCourseSerializer
from .course_group import InstitutionCourseGroupCourseSerializer
from .course_group import InstitutionCourseGroupSerializer
from .course_learner import InstitutionCourseLearnerSerializer
from .course_instructor import InstitutionCourseInstructorSerializer
from .informed_consent import InstitutionInformedConsentSerializer
from .informed_consent_document import InstitutionInformedConsentDocumentSerializer
from .institution import InstitutionSerializer
from .instructor import InstitutionInstructorSerializer
from .learner import InstitutionLearnerDetailSerializer
from .learner import InstitutionLearnerICBodySerializer
from .learner import InstitutionLearnerSerializer
from .request import InstitutionCourseActivityReportRequestSerializer
from .send_category import InstitutionSENDCategorySerializer
from .send_learner import InstitutionSENDLearnerSerializer
from .ui_option import InstitutionUIOptionSerializer
from .user import InstitutionUserSerializer
from .vle import InstitutionVLESerializer

__all__ = [
    "InstitutionSerializer",
    "InstitutionCourseSerializer",
    "InstitutionCourseActivitySerializer",
    "InstitutionCourseActivityExtendedSerializer",
    "InstitutionCourseActivityReportSerializer",
    "InstitutionCourseActivityReportAuditSerializer",
    "InstitutionCourseActivityReportDetailSerializer",
    "InstitutionCourseActivityReportExtendedSerializer",
    "InstitutionCourseActivityReportLearnerSerializer",
    "InstitutionCourseActivityInstrumentSerializer",
    "InstitutionCourseGroupSerializer",
    "InstitutionCourseGroupCourseSerializer",
    "InstitutionCourseLearnerSerializer",
    "InstitutionLearnerSerializer",
    "InstitutionCourseInstructorSerializer",
    "InstitutionCourseActivityReportRequestSerializer",
    "InstitutionLearnerDetailSerializer",
    "InstitutionInstructorSerializer",
    "InstitutionInformedConsentSerializer",
    "InstitutionLearnerICBodySerializer",
    "InstitutionInformedConsentDocumentSerializer",
    "InstitutionSENDCategorySerializer",
    "InstitutionSENDLearnerSerializer",
    "InstitutionVLESerializer",
    "InstitutionUIOptionSerializer",
    "InstitutionUserSerializer",
]



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
""" Configuration Manager Module"""
import configparser
import os

import hvac
from hvac.exceptions import InvalidRequest

from .conf import Config
from ..checks import check_vault_connection
from ..exception import TeslaConfigException


class ConfigManager:
    """ Manager class for configuration

        Contains the utility methods to manage configuration
    """
    #: Default path for secrets
    secrets_path = '/run/secrets'

    #: Search paths for configuration
    search_path = ['./tesla-ce.cfg', '/etc/tesla/tesla-ce.cfg']

    def __init__(self, load_config=True, secrets_path='/run/secrets', use_env=True, use_vault=True, config_file=None):
        """
            Default constructor

            :param load_config: Load configuration from default sources
            :type load_config: bool
            :param secrets_path: Base path for secrets. Default /run/secrets
            :type secrets_path: str
            :param use_env: Load configuration from environment variables
            :type use_env: bool
            :param use_vault: Load configuration from HashiCorp Vault
            :type use_vault: bool
            :param config_file: File to load the configuration from. If provided, other options will be ignored.
            :type config_file: str
        """
        #: Configuration object where configuration is stored
        self.config = Config()

        # Set secrets path
        self.secrets_path = secrets_path

        # List of enabled modules
        self.enabled_modules = []

        # Modules data
        self.modules = {}

        # Load configuration
        if load_config:
            # If a file is not provided, search in standard locations
            if config_file is None:
                config_file = self.find_config_file()

            # Load the configuration
            self.load(use_env=use_env, use_vault=use_vault, file_path=config_file)

    @classmethod
    def find_config_file(cls):
        """
            Looks for configuration path in the list of search paths

            :return: Path to the configuration file or None if not found
        """
        for file in cls.search_path:
            if os.path.exists(os.path.abspath(file)):
                return os.path.abspath(file)
        return None

    def save_configuration(self):
        """
            Save current configuration to the configuration file
        """
        # Check if configuration has been loaded from a file
        config_file = self.config.get('TESLA_CONFIG_FILE')

        # If there is no file for the configuration, find the most appropriate path
        if config_file is None:
            # TODO: Try with /etc/tesla first
            config_file = 'tesla-ce.cfg'

        # Write configuration to disk
        with open(config_file, 'w') as conf_fh:
            self.config.write(conf_fh)
        # Change permissions to the file
        os.chmod(config_file, 0o600)

    def load(self, file_path=None, vault_role=None, vault_secret=None, vault_url=None, use_env=True, use_vault=True):
        """
            Load configuration options

            :param file_path: If provided, read options in file.
            :type file_path: str
            :param vault_role: Role value to connect to Vault. If provided value in environment variable is not used
            :type vault_role: str
            :param vault_secret: Secret value to connect to Vault. If provided value in environment variable is not used
            :type vault_secret: str
            :param vault_url: Full URL to connect with vault (ie. https://vault:8200)
            :type vault_url: str
            :param use_env: Load configuration from environment variables
            :type use_env: bool
            :param use_vault: Load configuration from HashiCorp Vault
            :type use_vault: bool
            :exception TeslaConfigException: If the provided options are invalid
        """
        if not use_env and file_path is None and (vault_role is None or vault_secret is None):
            raise TeslaConfigException("Environment based config is disabled and no alternative is provided")

        # If file is provided, load configuration from file
        if file_path is not None:
            # Load configuration from file
            self.load_file(file_path)

            # Store the configuration file
            self.config.set('TESLA_CONFIG_FILE', file_path)
        # Try to load configuration from system
        if use_env:
            # Load configuration from secrets and environment variables
            self.load_env()
        # Get configuration from Vault
        if use_vault:
            if vault_role is not None and vault_secret is not None:
                # Update authentication values for vault with provided ones
                self.config.set('VAULT_ROLE_ID', vault_role)
                self.config.set('VAULT_SECRET_ID', vault_secret)
            if vault_url is not None:
                self.config.set('VAULT_URL', vault_url)
            if self.config.get('VAULT_ROLE_ID') is not None and self.config.get('VAULT_SECRET_ID') is not None:
                self.load_vault()

    def load_file(self, path):
        """
            Load configuration options from file

            :param path: Absolute path to configuration file
            :type path: str
            :exception TeslaConfigException: If the provided file is not found or is incorrect
        """
        if path is not None:
            if os.path.exists(path):
                conf = configparser.ConfigParser()
                with open(path, 'r') as conf_file:
                    conf.read_file(conf_file, path)
                    self.config.update(conf)
            else:
                raise TeslaConfigException("Cannot find provided file: {}".format(path))

    @staticmethod
    def _get_vault_entity_data(role_id, secret_id, vault_url, verify_ssl=None,
                               approle_path='tesla-ce/approle', kv_path='tesla-ce/kv'):
        """
            Get entity data from Vault.

            :param role_id: RoleID value. If not provided, current configuration will be used to find this value
            :type role_id: str
            :param secret_id: SecretID value. If not provided, current configuration will be used to find this value
            :type secret_id: str
            :param vault_url: Url to connect with Vault. If not provided, current configuration will be used to find
                this value
            :type vault_url: str
            :exception TeslaConfigException: If the provided arguments are invalid or vault is not accessible
            :param verify_ssl: Indicates if the certificate provided by Valut should be verified and the CA
            :type verify_ssl: str | bool
            :param approle_path: Mount point for AppRole authentication service
            :type approle_path: str
            :param kv_path: Mount point for KV secrets service
            :type kv_path: str
            :return: Entity data from Vault
            :rtype: dict
        """
        # Load configuration for each credential
        client = hvac.Client(url=vault_url, verify=verify_ssl)
        try:
            auth_data = client.auth.approle.login(role_id, secret_id, mount_point=approle_path)
        except InvalidRequest:
            raise TeslaConfigException("Cannot authenticate with provided credentials")
        entity_data = client.secrets.kv.v2.read_secret_version(
            path='modules/{}/metadata'.format(auth_data['auth']['metadata']['role_name']),
            mount_point=kv_path
        )['data']['data']['data']

        conf = client.secrets.kv.v2.read_secret_version(path=entity_data['config'],
                                                        mount_point=kv_path)['data']['data']
        apps = client.secrets.kv.v2.read_secret_version(path=entity_data['apps'],
                                                        mount_point=kv_path)['data']['data']
        services = client.secrets.kv.v2.read_secret_version(path=entity_data['services'],
                                                            mount_point=kv_path)['data']['data']

        # Create the module description structure
        module = {
            'module': entity_data['module'],
            'description': entity_data['description'],
            'apps': apps['data'],
            'services': services['data'],
            'config': {
                'VAULT_URL': vault_url,
                'VAULT_ROLE_ID': role_id,
                'VAULT_SECRET_ID': secret_id,
                'VAULT_SSL_VERIFY': verify_ssl,
                'VAULT_MOUNT_PATH_APPROLE': approle_path,
                'VAULT_MOUNT_PATH_KV': kv_path
            },
            'queues': entity_data.get('queues')
        }

        # Add retrieved values
        for keypair in conf['data']:
            key = keypair[0]
            path = keypair[1]
            data = client.secrets.kv.v2.read_secret_version(path=path, mount_point=kv_path)['data']['data']
            value = data['value']
            module['config'][key] = value

        return module

    def load_vault(self, role_id=None, secret_id=None, vault_url=None, approle_path=None, kv_path=None):
        """
            Load configuration options from Hashicorp Vault

            :param role_id: RoleID value. If not provided, current configuration will be used to find this value
            :type role_id: str | list
            :param secret_id: SecretID value. If not provided, current configuration will be used to find this value
            :type secret_id: str | list
            :param vault_url: Url to connect with Vault. If not provided, current configuration will be used to find
                this value
            :type vault_url: str
            :param approle_path: Mount point for AppRole authentication service
            :type approle_path: str
            :param kv_path: Mount point for KV secrets service
            :type kv_path: str
            :exception TeslaConfigException: If the provided arguments are invalid or vault is not accessible
        """
        if role_id is None:
            role_id = self.config.get('VAULT_ROLE_ID')
        else:
            self.config.set('VAULT_ROLE_ID', role_id)
        if secret_id is None:
            secret_id = self.config.get('VAULT_SECRET_ID')
        else:
            self.config.set('VAULT_SECRET_ID', secret_id)
        if approle_path is None:
            approle_path = self.config.get('VAULT_MOUNT_PATH_APPROLE')
        else:
            self.config.set('VAULT_MOUNT_PATH_APPROLE', approle_path)
        if kv_path is None:
            kv_path = self.config.get('VAULT_MOUNT_PATH_KV')
        else:
            self.config.set('VAULT_MOUNT_PATH_KV', kv_path)
        if vault_url is None:
            vault_url = self.config.get('VAULT_URL')
        else:
            self.config.set('VAULT_URL', vault_url)
        verify_ssl = self.config.get('VAULT_SSL_VERIFY')

        # Check if required information is provided
        if role_id is None or secret_id is None or vault_url is None:
            # Skip configuration
            return

        # Check if vault is available
        vault_connection_report = check_vault_connection(config=self.config)
        if not vault_connection_report['ready']:
            if role_id is not None and secret_id is not None:
                # Starting as module. Abort if cannot connect to Vault
                raise TeslaConfigException('Cannot connect to Vault with provided credentials.')
            # Skip configuration, as we are in Configuration mode
            return

        # Check if multiple credentials are provided
        self.enabled_modules = []
        if isinstance(role_id, list):
            assert len(role_id) == len(secret_id)
        else:
            role_id = [role_id]
            secret_id = [secret_id]
        for idx, role in enumerate(role_id):
            module = self._get_vault_entity_data(role_id[idx], secret_id[idx], vault_url, verify_ssl=verify_ssl,
                                                 approle_path=approle_path, kv_path=kv_path)
            self.modules[module['module']] = module
            self.enabled_modules.append(module['module'])

        # Get configuration from all modules
        config = self._get_config()
        for key in config:
            self.config.set(key, config[key])

    def load_env(self):
        """
            Load configuration options from Docker Secrets and Environment variables
        """
        # Read values in secrets
        if os.path.exists(self.secrets_path):
            for secret in list(os.scandir(self.secrets_path)):
                if self.config.is_valid_key(secret.name):
                    with open(secret.path, 'r') as f_secret:
                        value = f_secret.read()
                        self.config.set(secret.name, value)

        # Read options from environment variables
        for k in os.environ.keys():
            if self.config.is_valid_key(k):
                self.config.set(k, os.getenv(k))
            elif k.upper().endswith('_FILE'):
                # Try to read from file
                key = k.upper().split('_FILE')[0]
                if os.path.exists(os.getenv(k)) and self.config.is_valid_key(key):
                    with open(os.getenv(k), 'r') as f_secret:
                        value = f_secret.read()
                        self.config.set(key, value)

    def get_module(self, module=None):
        """
            Get the module information

            :param module: Name of the module in case of multiple modules enabled
            :type module: str
            :return: Module information
            :rtype: dict
        """
        if self.modules is None or len(self.modules) == 0:
            raise TeslaConfigException('No module configuration available')

        if module is None and len(self.modules) > 1:
            raise TeslaConfigException('Multiple modules enabled. Module must be provided')

        if module is None:
            module = list(self.modules.keys())[0]

        if module in self.modules.keys():
            return self.modules[module]

        raise TeslaConfigException('No module configuration available for {}'.format(module))

    def get_module_config(self, module=None):
        """
            Get the module configuration

            :param module: Name of the module in case of multiple modules enabled
            :type module: str
            :return: Module configuration
            :rtype: dict
        """
        info = self.get_module(module)
        if 'config' in info:
            return info['config']

        raise TeslaConfigException('No module configuration available')

    def _get_config(self):
        """
            Get configuration object merging all available modules configurations

            :return: Module configuration
            :rtype: dict
        """
        has_api = False
        config = {}
        for module in self.modules:
            if module == 'api':
                has_api = True
                continue
            mod_config = self.get_module_config(module)
            config.update(mod_config)
        # API configuration have precedence. If api is one of the modules, apply it the end
        if has_api:
            api_config = self.get_module_config('api')
            config.update(api_config)

        return config

    def get_database_config(self):
        """
            Return database configurations

            :return: Database configurations
            :rtype: dict
        """
        db = {}
        has_db = False
        for module in self.modules:
            if self.modules[module].get('services') is not None and 'db' in self.modules[module].get('services'):
                has_db = True
                mod_config = self.get_module_config(module)
                db[module] = {
                    'ENGINE': 'django.db.backends.{}'.format(mod_config.get('DB_ENGINE')),
                    'HOST': mod_config.get('DB_HOST'),
                    'NAME': mod_config.get('DB_NAME'),
                    'USER': mod_config.get('DB_USER'),
                    'PASSWORD': mod_config.get('DB_PASSWORD'),
                    'PORT': mod_config.get('DB_PORT'),
                }
        if len(list(db.keys())) > 0:
            db['default'] = db[list(db.keys())[0]]
        if not has_db:
            db = {}
        return db

    def get_cache_config(self):
        """
            Return cache configuration

            :return: Cache configuration
            :rtype: dict
        """
        cache = {}
        has_cache = False
        for module in self.modules:
            if self.modules[module].get('services') is not None and 'redis' in self.modules[module].get('services'):
                has_cache = True
                mod_config = self.get_module_config(module)
                cache[module] = {
                    #'BACKEND': 'django_redis.cache.RedisCache',
                    'BACKEND': 'django_prometheus.cache.backends.redis.RedisCache',
                    'LOCATION': "redis://{}:{}/{}".format(mod_config.get('REDIS_HOST'),
                                                          mod_config.get('REDIS_PORT'),
                                                          mod_config.get('REDIS_DATABASE')),
                    'OPTIONS': {
                        'CLIENT_CLASS': 'django_redis.client.DefaultClient',
                        'PASSWORD': mod_config.get('REDIS_PASSWORD')
                    }
                }
        if len(list(cache.keys())) > 0:
            cache['default'] = cache[list(cache.keys())[0]]
        if not has_cache:
            cache['default'] = {
                #'BACKEND': 'django_redis.cache.RedisCache',
                'BACKEND': 'django_prometheus.cache.backends.redis.RedisCache',
                'LOCATION': "redis://{}:{}/{}".format(self.config.get('REDIS_HOST'),
                                                      self.config.get('REDIS_PORT'),
                                                      self.config.get('REDIS_DATABASE')),
                'OPTIONS': {
                    'CLIENT_CLASS': 'django_redis.client.DefaultClient',
                    'PASSWORD': self.config.get('REDIS_PASSWORD')
                }
            }
        return cache

    def get_celery_config(self):
        """
            Return celery broker url

            :return: Brocker url for each module
            :rtype: str
        """
        celery = {}
        has_celery = False
        for module in self.modules:
            if self.modules[module].get('services') is not None and 'celery' in self.modules[module].get('services'):
                has_celery = True
                mod_config = self.get_module_config(module)
                if mod_config.get('CELERY_BROKER_PROTOCOL') == 'sqs':
                    celery[module] = {
                        'CELERY_BROKER_TRANSPORT_OPTIONS': {'region': mod_config.get('CELERY_BROKER_REGION')},
                        'CELERY_BROKER_URL': '{0}://{1}:{2}@'.format(
                            'sqs',
                            mod_config.get('CELERY_BROKER_USER'),
                            mod_config.get('CELERY_BROKER_PASSWORD')
                        )
                    }
                else:
                    vhost = mod_config.get('CELERY_BROKER_VHOST')
                    if len(vhost) == 0 or vhost is None:
                        vhost = '/'
                    elif not vhost.startswith('/'):
                        vhost = '/' + vhost
                    celery[module] = {
                        'CELERY_BROKER_URL': '{}://{}:{}@{}:{}{}'.format(
                            mod_config.get('CELERY_BROKER_PROTOCOL'),
                            mod_config.get('CELERY_BROKER_USER'),
                            mod_config.get('CELERY_BROKER_PASSWORD'),
                            mod_config.get('CELERY_BROKER_HOST'),
                            mod_config.get('CELERY_BROKER_PORT'),
                            vhost
                        )
                    }

        if len(list(celery.keys())) > 0:
            if len(list(celery.keys())) > 1:
                if self.config.get('RABBITMQ_ADMIN_USER') is not None and \
                        self.config.get('RABBITMQ_ADMIN_PASSWORD') is not None:
                    celery = celery[list(celery.keys())[0]]['CELERY_BROKER_URL']
                    protocol = celery.split('//')[0]
                    host = celery.split('@')[1]
                    celery = '{}//{}:{}@{}'.format(protocol,
                                                   self.config.get('RABBITMQ_ADMIN_USER'),
                                                   self.config.get('RABBITMQ_ADMIN_PASSWORD'),
                                                   host
                                                   )
                else:
                    raise TeslaConfigException('Multiple celery configuration without administration credentials.')
            else:
                celery = celery[list(celery.keys())[0]]['CELERY_BROKER_URL']
        if not has_celery:
            # Use default parameters
            vhost = self.config.get('CELERY_BROKER_VHOST')
            if len(vhost) == 0 or vhost is None:
                vhost = '/'
            elif not vhost.startswith('/'):
                vhost = '/' + vhost
            celery = '{}://{}:{}@{}:{}{}'.format(
                        self.config.get('CELERY_BROKER_PROTOCOL'),
                        self.config.get('CELERY_BROKER_USER'),
                        self.config.get('CELERY_BROKER_PASSWORD'),
                        self.config.get('CELERY_BROKER_HOST'),
                        self.config.get('CELERY_BROKER_PORT'),
                        vhost
                    )

        return celery

    def get_status(self):
        """
            Get the configuration status

            :return: Object with status information
            :rtype: dict
        """
        return {
            'status': 0,
            'warnings': 0,
            'errors': 0,
            'info': {}
        }



# Generated by Django 3.0 on 2020-09-30 16:29

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('tesla_ce', '0009_monitor'),
    ]

    operations = [
        migrations.AddField(
            model_name='institutionuser',
            name='inst_admin',
            field=models.BooleanField(default=False, help_text='Whether this user is administrator of the institution'),
        ),
        migrations.AddField(
            model_name='institutionuser',
            name='legal_admin',
            field=models.BooleanField(default=False, help_text='Whether this user can manage legal data of the institution'),
        ),
        migrations.AddField(
            model_name='institutionuser',
            name='send_admin',
            field=models.BooleanField(default=False, help_text='Whether this user can manage SEND data of the institution'),
        ),
    ]



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Instructor api serialize module."""
from rest_framework import serializers
from rest_framework import validators

from tesla_ce.models import Institution
from tesla_ce.models import InstitutionUser
from tesla_ce.models import Instructor
from tesla_ce.models import User


class InstitutionCourseInstructorSerializer(serializers.ModelSerializer):
    """Instructor serialize model module."""
    username = serializers.CharField(read_only=True)
    uid = serializers.CharField()
    first_name = serializers.CharField(read_only=True)
    last_name = serializers.CharField(read_only=True)
    email = serializers.EmailField(read_only=True)
    institution_id = serializers.HiddenField(allow_null=True, default=None)
    course_id = serializers.HiddenField(allow_null=True, default=None)

    class Meta:
        model = Instructor
        exclude = ["password", "is_superuser", "is_staff", "is_active", "date_joined", "groups",
                   "user_permissions", "institution", "login_allowed", "inst_admin", "send_admin",
                   "data_admin", "legal_admin", "last_login", "locale"]

    def validate(self, attrs):
        """
            Validate the given attributes
            :param attrs: Attributes parsed from request
            :type attrs: dict
            :return: Validated attributes
            :rtype: dict
        """
        # Add predefined fields
        attrs['institution_id'] = self.context['view'].kwargs['parent_lookup_vle__institution_id']
        attrs['course_id'] = self.context['view'].kwargs['parent_lookup_id']

        # Apply validators
        for validator in self.get_validators():
            validator(attrs, self)
        return super().validate(attrs)



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Learner enrolment samples views module """
import json

import botocore
from django.core.files.base import ContentFile
from django.core.files.storage import default_storage
from rest_framework.views import APIView
from rest_framework.views import Response
from rest_framework.views import status

from tesla_ce.lib.exception import tesla_report_exception
from ..utils import is_authenticated


class BaseLearnerAPIView(APIView):
    """
        Base class for Learner API views
    """
    serializer_class = None

    def get_storing_path(self, institution_id, data):
        """
            Create the path where this sample should be stored
            :param institution_id: Id of the institution
            :type institution_id: int
            :param data: The data of the request
            :type data: dict
            :return: Path to the storage
            :rtype: str
        """
        raise NotImplementedError('This method must be implemented')

    def enqueue_request(self, data, path):
        """
            Enqueue a new task to manage new storage data
            :param data: Stored data
            :type data: dict
            :param path: Path where the data has been stored
            :type path: str
        """
        raise NotImplementedError('This method must be implemented')

    def post(self, request, institution_id, learner_id, format=None):
        """
            Perform data storage
            :param request: The received request
            :param institution_id: Institution unique ID
            :param learner_id: Learner ID in TeSLA
            :param format: Format of the request
        """
        serializer = self.serializer_class(data=request.data, context={'request': request})
        try:
            if serializer.is_valid():
                # Check that provided token is compatible with the data
                if not is_authenticated(request, institution_id, learner_id, serializer.data):
                    return Response(status=status.HTTP_403_FORBIDDEN, data={'status': "ERROR", 'errors': "Invalid token"})
                # Generate the target path for the request
                path = self.get_storing_path(institution_id, serializer.data)
                # Store the request
                resp = default_storage.save(path, ContentFile(json.dumps(serializer.data).encode('utf-8')))
                # Notify the new stored request
                self.enqueue_request(serializer.data, resp)

                return Response({"status": "OK", "path": resp})

            return Response(status=status.HTTP_400_BAD_REQUEST,
                            data={"status": "ERROR", 'errors': serializer.errors},
                            content_type="application/json")
        except botocore.exceptions.ClientError as cli_err:
            tesla_report_exception(cli_err)
            return Response(status=status.HTTP_500_INTERNAL_SERVER_ERROR,
                            data={"status": "ERROR", 'errors': cli_err.response},
                            content_type="application/json")

        except Exception as ex:
            tesla_report_exception(ex)
            return Response(status=status.HTTP_500_INTERNAL_SERVER_ERROR,
                            data={"status": "ERROR", 'errors': ex},
                            content_type="application/json")



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
""" Storage Manager Module"""
from botocore import exceptions
from django.core.management import call_command
from ..config import ConfigManager
from ..exception import TeslaStorageException


class StorageManager:
    """ Manager class for Storage """

    #: Storage url
    storage_url: str = None

    #: Storage client
    _client = None

    def __init__(self, config=None):
        """
            Default constructor
            :param config: Configuration manager instance
            :type config: ConfigManager
        """
        if config is None:
            config = ConfigManager()

        # Store the configuration object
        self._config = config

    @property
    def client(self):
        """
            Access to the Storage server
            :return: Storage server object
            :rtype: s3boto3.S3Boto3Storage
        """
        if self._client is None:
            # Import when client is created to avoid errors during installation process
            from storages.backends import s3boto3
            self._client = s3boto3.S3Boto3Storage()
        return self._client

    def initialize(self):
        """
            Initialize the storage
        """
        # Get the list of buckets
        buckets = [b.name for b in self._client.connection.buckets.all()]

        # Create the bucket if it does not exist
        if self._config.config.get('STORAGE_BUCKET_NAME') not in buckets:
            self.create_default_bucket()
        if self._config.config.get('STORAGE_PUBLIC_BUCKET_NAME') not in buckets:
            self.create_public_bucket()

        # Export static files
        call_command('collectstatic', verbosity=0, interactive=False)

    def create_default_bucket(self):
        """
            Create the default bucket in the configuration
        """
        # Get configuration
        bucket = self._config.config.get('STORAGE_BUCKET_NAME')
        region = self._config.config.get('STORAGE_REGION')
        # Create bucket
        try:
            if region is None:
                self._client.connection.create_bucket(Bucket=bucket, ACL='private')
            else:
                self._client.connection.create_bucket(Bucket=bucket, ACL='private',
                                                      CreateBucketConfiguration={'LocationConstraint': region})
        except exceptions.ClientError as e:
            raise TeslaStorageException('Cannot create bucket {} at region {}: {}'.format(bucket, region, str(e)))

    def create_public_bucket(self):
        """
            Create the public bucket in the configuration
        """
        # Get configuration
        bucket = self._config.config.get('STORAGE_PUBLIC_BUCKET_NAME')
        region = self._config.config.get('STORAGE_REGION')
        # Create bucket
        try:
            if region is None:
                self._client.connection.create_bucket(Bucket=bucket, ACL='download')
            else:
                self._client.connection.create_bucket(Bucket=bucket, ACL='download',
                                                      CreateBucketConfiguration={'LocationConstraint': region})
        except exceptions.ClientError as e:
            raise TeslaStorageException('Cannot create bucket {} at region {}: {}'.format(bucket, region, str(e)))

    def get_status(self):
        """
            Get the storage status

            :return: Object with status information
            :rtype: dict
        """
        errors = []
        warnings = []

        can_connect = False
        ready = False
        try:
            self.client.listdir('/')
            can_connect = True
            ready = True
        except exceptions.ClientError as e:
            if e.response['Error']['Code'] == "NoSuchBucket":
                can_connect = True
                warnings.append('Bucket {} does not exist'.format(self.client.bucket_name))
            elif e.response['Error']['Code'] == "InvalidAccessKeyId" or \
                    e.response['Error']['Code'] == "SignatureDoesNotMatch":
                errors.append('Invalid credentials')
            else:
                errors.append('Could not connect to URL {}'.format(self.client.endpoint_url))
        except exceptions.EndpointConnectionError:
            errors.append('Could not connect to URL {}'.format(self.client.endpoint_url))
        except exceptions.SSLError:
            errors.append('Invalid SSL certificate connecting to URL {}'.format(self.client.endpoint_url))
        except Exception as e:
            errors.append('Error connecting to storage server: {}'.format(str(e)))
        status = 1
        if not can_connect:
            status = 0
        return {
            'status': status,
            'warnings': len(warnings),
            'errors': len(errors),
            'info': {
                'ready': ready,
                'errors': errors,
                'warnings': warnings,
            }
        }



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""User api serialize module."""
from django.contrib.auth.models import make_password

from rest_framework import serializers
from rest_framework import validators

from tesla_ce import get_default_client

from tesla_ce.models import InstitutionUser
from tesla_ce.models import User


class InstitutionUserSerializer(serializers.ModelSerializer):
    """Institution User serialize model module."""

    username = serializers.CharField(allow_null=False,
                                     allow_blank=False,
                                     validators=[
                                         validators.UniqueValidator(queryset=User.objects.all())
                                     ]
                                     )
    password = serializers.CharField(write_only=True, default=None, allow_null=True)
    password2 = serializers.CharField(write_only=True, default=None, allow_null=True)
    last_login = serializers.DateTimeField(read_only=True)
    first_name = serializers.CharField(required=False, default=None)
    last_name = serializers.CharField(required=False, default=None)
    uid = serializers.CharField(required=True, allow_null=False, allow_blank=False)
    email = serializers.EmailField(required=False, default=None,
                                   validators=[
                                       validators.UniqueValidator(queryset=User.objects.all())
                                   ]
                                   )
    institution_id = serializers.HiddenField(default=None)
    login_allowed = serializers.BooleanField(required=False, allow_null=False, default=False)
    inst_admin = serializers.BooleanField(required=False, allow_null=False, default=False)
    legal_admin = serializers.BooleanField(required=False, allow_null=False, default=False)
    send_admin = serializers.BooleanField(required=False, allow_null=False, default=False)
    data_admin = serializers.BooleanField(required=False, allow_null=False, default=False)

    class Meta:
        model = InstitutionUser
        exclude = ["is_superuser", "is_staff", "is_active", "date_joined", "groups",
                   "user_permissions", "institution"]
        validators = [serializers.UniqueTogetherValidator(
            queryset=InstitutionUser.objects.all(),
            fields=['institution_id', 'uid']
        ), ]

    def validate(self, attrs):
        """
            Validate the given attributes
            :param attrs: Attributes parsed from request
            :type attrs: dict
            :return: Validated attributes
            :rtype: dict
        """
        # Add predefined fields
        attrs['institution_id'] = self.context['view'].kwargs['parent_lookup_institution_id']

        # Check passwords
        if attrs.get('login_allowed', True):
            if 'password' in attrs and attrs.get('password') != attrs.get('password2'):
                raise serializers.ValidationError('Passwords does not match')
            if 'password2' in attrs:
                del attrs['password2']
        else:
            attrs['login_allowed'] = False

        # Apply validators
        for validator in self.get_validators():
            validator(attrs, self)
        return super().validate(attrs)

    def create(self, validated_data):
        new_user = super().create(validated_data)

        if 'password' in validated_data:
            if new_user.login_allowed:
                get_default_client().change_user_password(new_user.email, validated_data['password'])
            else:
                new_user.set_unusable_password()

        return new_user

    def update(self, instance, validated_data):
        updated_user = super().update(instance, validated_data)

        if 'password' in validated_data:
            if updated_user.login_allowed:
                get_default_client().change_user_password(updated_user.email, validated_data['password'])
            else:
                updated_user.set_unusable_password()

        return updated_user




#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Activity Report api serialize module."""
from rest_framework import serializers

from tesla_ce.models import Learner
from tesla_ce.models import ReportActivity
from tesla_ce.models import ReportActivityInstrument


class VLECourseActivityReportLearnerSerializer(serializers.ModelSerializer):
    """Activity Report Learner serialize model module."""
    class Meta:
        model = Learner
        fields = ["id", "uid", "first_name", "last_name", "email", "learner_id"]


class VLECourseActivityReportDetailSerializer(serializers.ModelSerializer):
    """Activity Report Detail serialize model module."""
    instrument = serializers.CharField(source='instrument.name')
    instrument_id = serializers.IntegerField(source='instrument.id')
    enrolment = serializers.FloatField()
    confidence = serializers.FloatField()
    result = serializers.FloatField()

    class Meta:
        model = ReportActivityInstrument
        fields = ["instrument", "enrolment", "confidence", "result", "identity_level",
                  "content_level", "integrity_level", "instrument_id"]


class VLECourseActivityReportExtendedDetailSerializer(serializers.ModelSerializer):
    """Activity Report Detail serialize model module."""
    instrument = serializers.CharField(source='instrument.name')
    instrument_id = serializers.IntegerField(source='instrument.id')
    instrument_acronym = serializers.CharField(source='instrument.acronym')
    instrument_polarity = serializers.SerializerMethodField()
    enrolment = serializers.FloatField()
    confidence = serializers.FloatField()
    result = serializers.FloatField()
    learner_histogram = serializers.SerializerMethodField()
    activity_histogram = serializers.SerializerMethodField()
    result_bean = serializers.SerializerMethodField()
    thresholds = serializers.SerializerMethodField()
    prob_learner = serializers.SerializerMethodField()
    prob_context = serializers.SerializerMethodField()
    h_prob_learner = serializers.SerializerMethodField()
    h_prob_context = serializers.SerializerMethodField()

    class Meta:
        model = ReportActivityInstrument
        fields = ["instrument", "enrolment", "confidence", "result", "identity_level", "instrument_acronym",
                  "content_level", "integrity_level", "learner_histogram", "activity_histogram",
                  "instrument_id", "instrument_polarity", "result_bean", "thresholds",
                  "prob_learner", "prob_context", "h_prob_learner", "h_prob_context"]

    def get_activity_histogram(self, instance):
        return list(instance.instrument.histogramactivityinstrument_set.values_list(
            'b0', 'b1', 'b2', 'b3', 'b4', 'b5', 'b6', 'b7', 'b8', 'b9').first())

    def get_learner_histogram(self, instance):
        return list(instance.report.learner.histogramlearnerinstrument_set.filter(
            instrument=instance.instrument
        ).values_list('b0', 'b1', 'b2', 'b3', 'b4', 'b5', 'b6', 'b7', 'b8', 'b9').first())

    def get_instrument_polarity(self, instance):
        return 2 * int(not instance.instrument.provider_set.first().inverted_polarity) - 1

    def get_result_bean(self, instance):
        return min(int(instance.result/10), 9)

    def get_thresholds(self, instance):
        prov_thr = instance.instrument.provider_set.first()
        return {
            "warning_below": prov_thr.warning_below,
            "alert_below": prov_thr.alert_below
        }

    def get_prob_learner(self, instance):
        bin = self.get_result_bean(instance)
        hist = self.get_learner_histogram(instance)
        acc = hist[bin]
        if bin > 0:
            acc += hist[bin - 1] / 2.0
        if bin < 9:
            acc += hist[bin + 1] / 2.0

        return acc / sum(hist)

    def get_prob_context(self, instance):
        bin = self.get_result_bean(instance)
        hist = self.get_activity_histogram(instance)
        acc = hist[bin]
        if bin > 0:
            acc += hist[bin - 1] / 2.0
        if bin < 9:
            acc += hist[bin + 1] / 2.0

        return acc / sum(hist)

    def get_h_prob_learner(self, instance):
        bin = self.get_result_bean(instance)
        hist = self.get_learner_histogram(instance)
        if self.get_instrument_polarity(instance) > 0:
            acc = sum(hist[bin + 1: 10])
        else:
            acc = sum(hist[0:bin])

        return acc / sum(hist)

    def get_h_prob_context(self, instance):
        bin = self.get_result_bean(instance)
        hist = self.get_activity_histogram(instance)
        if self.get_instrument_polarity(instance) > 0:
            acc = sum(hist[bin + 1: 10])
        else:
            acc = sum(hist[0:bin])

        return acc / sum(hist)


class VLECourseActivityReportSerializer(serializers.ModelSerializer):
    """Activity Report serialize model module."""
    learner = VLECourseActivityReportLearnerSerializer()
    detail = VLECourseActivityReportDetailSerializer(source='reportactivityinstrument_set', many=True)

    class Meta:
        model = ReportActivity
        fields = ["id", "learner", "detail", "identity_level", "content_level", "integrity_level"]


class VLECourseActivityReportExtendedSerializer(serializers.ModelSerializer):
    """Activity Report serialize model module."""
    learner = VLECourseActivityReportLearnerSerializer()
    detail = VLECourseActivityReportExtendedDetailSerializer(source='reportactivityinstrument_set', many=True)

    class Meta:
        model = ReportActivity
        fields = ["id", "learner", "detail", "identity_level", "content_level", "integrity_level"]



#  Copyright (c) 2021 Mireia Bellot
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Test module for institution users management """
import pytest

from tests import auth_utils
from tests.conftest import get_random_string


def test_api_institution_users(rest_api_client, institution_course_test_case):
    """
        CRUD for institution users
    """
    # Set institution admin user.
    inst_admin = institution_course_test_case['user']
    inst_admin.institutionuser.inst_admin = True
    inst_admin.institutionuser.save()

    # Get the instution id
    inst_id = institution_course_test_case['institution'].id

    # Authenticate with institution admin
    client = auth_utils.client_with_user_obj(inst_admin)

    # Create a new user
    user_name = get_random_string(10)
    password = get_random_string(10)
    email = '{}@inst.tesla-ce.eu'.format(user_name)
    user_data = {
        'username': user_name,
        'uid': user_name,
        'email': email,
        'first_name': user_name[:5],
        'last_name': user_name[5:],
        'login_allowed': True,
        'is_active': True,
        'is_staff': False,
        'password': password,
        'password2': password
    }
    user_create_resp = client.post('/api/v2/institution/{}/user/'.format(inst_id), data=user_data, format='json')
    assert user_create_resp.status_code == 201

    # Ensure that this user can authenticate
    user_client = auth_utils.client_with_user_credentials(email, password)
    profile = auth_utils.get_profile(user_client)
    assert profile['institution'] is not None
    assert profile['institution']['id'] == inst_id
    assert len(profile['roles']) == 0
    assert profile['username'] == user_name
    assert profile['email'] == email
    assert profile['first_name'] == user_name[:5]
    assert profile['last_name'] == user_name[5:]

    # Grant administration privileges
    user_mod_resp = client.patch(
        '/api/v2/institution/{}/user/{}/'.format(inst_id, user_create_resp.data['id']),
        data={'inst_admin': True},
        format='json'
    )
    assert user_mod_resp.status_code == 200

    # Get the user profile again
    profile = auth_utils.get_profile(user_client)
    assert len(profile['roles']) == 1
    assert 'ADMIN' in profile['roles']

    # Try to grant global administration privileges
    user_mod_resp2 = client.patch(
        '/api/v2/institution/{}/user/{}/'.format(inst_id, user_create_resp.data['id']),
        data={'is_staff': True},
        format='json'
    )
    assert user_mod_resp2.status_code == 200

    # Get the user profile again
    profile = auth_utils.get_profile(user_client)
    assert len(profile['roles']) == 1
    assert 'ADMIN' in profile['roles']

    # Find for this user by email
    find_user_resp = client.get('/api/v2/institution/{}/user/?email={}'.format(inst_id, email))
    assert find_user_resp.status_code == 200
    assert find_user_resp.data['count'] == 1
    assert find_user_resp.data['results'][0]['id'] == profile['id']

    # Remove the user
    user_del_resp = client.delete(
        '/api/v2/institution/{}/user/{}/'.format(inst_id, user_create_resp.data['id'])
    )
    assert user_del_resp.status_code == 204

    # Ensure that this user does not exist
    find2_user_resp = client.get('/api/v2/institution/{}/user/?email={}'.format(inst_id, email))
    assert find2_user_resp.status_code == 200
    assert find2_user_resp.data['count'] == 0


@pytest.mark.django_db
def test_api_institution_user_filters(rest_api_client, institution_course_test_case):
    """
        Test filters for getting the users
    """
    # Set institution admin user.
    inst_admin = institution_course_test_case['user']
    inst_admin.institutionuser.inst_admin = True
    inst_admin.institutionuser.save()

    # Get the instution id
    inst_id = institution_course_test_case['institution'].id

    # Authenticate with institution admin
    client = auth_utils.client_with_user_obj(inst_admin)

    # Get the admin user by email
    get_user_by_email_resp = client.get('/api/v2/institution/{}/user/?email={}'.format(inst_id, inst_admin.email))
    assert get_user_by_email_resp.status_code == 200
    assert get_user_by_email_resp.data['count'] == 1
    assert get_user_by_email_resp.data['results'][0]['id'] == inst_admin.id

    # Filter users using one role
    get_user_by_role_resp = client.get('/api/v2/institution/{}/user/?roles=ADMIN'.format(inst_id))
    assert get_user_by_role_resp.status_code == 200
    assert get_user_by_role_resp.data['count'] == 1
    assert get_user_by_role_resp.data['results'][0]['id'] == inst_admin.id

    # Filter users using multiple roles
    inst_admin.institutionuser.legal_admin = True
    inst_admin.institutionuser.save()
    get_user_by_role2_resp = client.get('/api/v2/institution/{}/user/?roles=ADMIN&roles=LEGAL'.format(inst_id))
    assert get_user_by_role2_resp.status_code == 200
    assert get_user_by_role2_resp.data['count'] == 1
    assert get_user_by_role2_resp.data['results'][0]['id'] == inst_admin.id

    # Filter learners
    get_user_by_learner_role_resp = client.get('/api/v2/institution/{}/user/?roles=LEARNER'.format(inst_id))
    assert get_user_by_learner_role_resp.status_code == 200
    assert get_user_by_learner_role_resp.data['count'] == 1
    assert get_user_by_learner_role_resp.data['results'][0]['id'] == institution_course_test_case['learner'].id

    # Filter instructors
    get_user_by_instructor_role_resp = client.get('/api/v2/institution/{}/user/?roles=INSTRUCTOR'.format(inst_id))
    assert get_user_by_instructor_role_resp.status_code == 200
    assert get_user_by_instructor_role_resp.data['count'] == 1
    assert get_user_by_instructor_role_resp.data['results'][0]['id'] == institution_course_test_case['instructor'].id

    # Find for any admin
    inst_admin = institution_course_test_case['user']
    inst_admin.institutionuser.inst_admin = True
    inst_admin.institutionuser.send_admin = True
    inst_admin.institutionuser.data_admin = True
    inst_admin.institutionuser.legal_admin = True
    inst_admin.institutionuser.save()

    get_user_by_admins_roles_resp = client.get(
        '/api/v2/institution/{}/user/?roles=ADMIN&roles=LEGAL&roles=DATA&roles=SEND'.format(inst_id)
    )
    assert get_user_by_admins_roles_resp.status_code == 200
    assert get_user_by_admins_roles_resp.data['count'] == 1
    assert get_user_by_admins_roles_resp.data['results'][0]['id'] == inst_admin.id



#  Copyright (c) 2021 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Module for Authentication API profile """
import pytest
import tests
from tests.utils import get_profile


@pytest.mark.django_db
def test_api_admin_ui_routes(rest_api_client, api_client, institution_course_test_case,
                             user_global_admin, empty_ui_routes):

    # Ensure there is no option defined
    rest_api_client.force_authenticate(user=user_global_admin)
    options_list = tests.utils.get_rest_api_client(rest_api_client, '/api/v2/admin/ui/',
                                                   'List ui options', 'RESPONSE:', 200)
    assert options_list['count'] == 0

    # Add a basic route for all users
    options_add = rest_api_client.post('/api/v2/admin/ui/', format='json', data={
        'enabled': True,
        'route': '/dashboard'
    })
    assert options_add.status_code == 201

    # Add an option for admins
    options_add_admin = rest_api_client.post('/api/v2/admin/ui/', format='json', data={
        'enabled': True,
        'route': '/admin',
        'roles': 'GLOBAL_ADMIN'
    })
    assert options_add_admin.status_code == 201

    options_list = tests.utils.get_rest_api_client(rest_api_client, '/api/v2/admin/ui/',
                                                   'List ui options', 'RESPONSE:', 200)
    assert options_list['count'] == 2
    assert options_list['results'][0]['route'] == '/dashboard' or options_list['results'][0]['route'] == '/admin'
    if options_list['results'][0]['route'] == '/admin':
        assert options_list['results'][1]['roles'] is None
        assert options_list['results'][0]['roles'] == 'GLOBAL_ADMIN'
    else:
        assert options_list['results'][0]['roles'] is None
        assert options_list['results'][1]['roles'] == 'GLOBAL_ADMIN'

    # Check profile routes
    # Global Admin
    global_admin_token = api_client.get_admin_token_pair(user_global_admin, {})
    global_admin_profile_resp = get_profile(rest_api_client, global_admin_token['access_token'])
    assert global_admin_profile_resp.status_code == 200
    assert 'routes' in global_admin_profile_resp.data
    assert len(global_admin_profile_resp.data['routes']) == 2
    assert '/admin' in global_admin_profile_resp.data['routes']
    assert '/dashboard' in global_admin_profile_resp.data['routes']

    # Institution user
    inst_user_token = api_client.get_user_token_pair(institution_course_test_case['user'].institutionuser, {})
    inst_user_profile_resp = get_profile(rest_api_client, inst_user_token['access_token'])
    assert inst_user_profile_resp.status_code == 200
    assert 'routes' in inst_user_profile_resp.data
    assert len(inst_user_profile_resp.data['routes']) == 1
    assert '/dashboard' in inst_user_profile_resp.data['routes']

    # Add learners and instructors routes
    options_add_courses = rest_api_client.post('/api/v2/admin/ui/', format='json', data={
        'enabled': True,
        'route': '/mycourses',
        'roles': 'LEARNER,INSTRUCTOR'
    })
    assert options_add_courses.status_code == 201

    # Check that new route is not included for global admins
    global_admin_profile_resp = get_profile(rest_api_client, global_admin_token['access_token'])
    assert global_admin_profile_resp.status_code == 200
    assert len(global_admin_profile_resp.data['routes']) == 2
    assert '/mycourses' not in global_admin_profile_resp.data['routes']
    assert '/admin' in global_admin_profile_resp.data['routes']
    assert '/dashboard' in global_admin_profile_resp.data['routes']

    # Check that new route is not included for institution admins
    inst_user_token = api_client.get_user_token_pair(institution_course_test_case['user'].institutionuser, {})
    inst_user_profile_resp = get_profile(rest_api_client, inst_user_token['access_token'])
    assert inst_user_profile_resp.status_code == 200
    assert len(inst_user_profile_resp.data['routes']) == 1
    assert '/mycourses' not in inst_user_profile_resp.data['routes']
    assert '/dashboard' in inst_user_profile_resp.data['routes']

    # Check that is included for learners
    learner_user = institution_course_test_case['learner'].institutionuser.learner
    inst_learner_user_token = api_client.get_user_token_pair(learner_user, {})
    inst_learner_user_profile_resp = get_profile(rest_api_client, inst_learner_user_token['access_token'])

    assert inst_learner_user_profile_resp.status_code == 200
    assert 'routes' in inst_learner_user_profile_resp.data
    assert len(inst_learner_user_profile_resp.data['routes']) == 2
    assert '/dashboard' in inst_learner_user_profile_resp.data['routes']
    assert '/mycourses' in inst_learner_user_profile_resp.data['routes']

    # Check that is included for instructors
    instructor_user = institution_course_test_case['instructor'].institutionuser.instructor
    inst_instructor_user_token = api_client.get_user_token_pair(instructor_user, {})
    inst_instructor_user_profile_resp = get_profile(rest_api_client, inst_instructor_user_token['access_token'])

    assert inst_instructor_user_profile_resp.status_code == 200
    assert 'routes' in inst_instructor_user_profile_resp.data
    assert len(inst_instructor_user_profile_resp.data['routes']) == 2
    assert '/dashboard' in inst_instructor_user_profile_resp.data['routes']
    assert '/mycourses' in inst_instructor_user_profile_resp.data['routes']



#  Copyright (c) 2021 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" TeSLA CE Learner actions for Use Case tests """
import mock
import requests

from django.utils import timezone

from tests import auth_utils
from tests.conftest import get_random_string

# List of submitted samples
samples_list = {}

# List of submitted requests
requests_list = {}


def api_learner_accept_ic(launcher):
    """
        The learner accepts the IC using the API
        :param launcher: Launcher object
    """
    # Authenticate with learner launcher credentials
    client = auth_utils.client_with_launcher_credentials(launcher)

    # Get the user profile
    profile = auth_utils.get_profile(client)
    assert "LEARNER" in profile['roles']

    # Get required data
    institution_id = profile['institution']['id']

    # Get the current IC
    get_current_ic_resp = client.get('/api/v2/institution/{}/ic/current/'.format(institution_id))
    assert get_current_ic_resp.status_code == 200

    # Accept the informed consent
    accept_ic_resp = client.post('/api/v2/institution/{}/learner/{}/ic/'.format(institution_id, profile['id']),
                                 data={'version': get_current_ic_resp.data['version']},
                                 format='json')
    assert accept_ic_resp.status_code == 200


def api_learner_enrolment(launcher):
    """
        A learner check their enrolment status via API.
        :param launcher: Launcher object
        :return: List of enrolments
    """
    # Authenticate with learner launcher credentials
    client = auth_utils.client_with_launcher_credentials(launcher)

    # Get the user profile
    profile = auth_utils.get_profile(client)
    assert "LEARNER" in profile['roles']

    # Get the list of enrolments
    institution_id = profile['institution']['id']
    list_enrolments_resp = client.get(
        '/api/v2/institution/{}/learner/{}/enrolment/'.format(institution_id, profile['id'])
    )
    assert list_enrolments_resp.status_code == 200

    return list_enrolments_resp.data


def api_learner_missing_enrolment(launcher, activity, missing=False):
    """
        A learner check missing enrolments for an activity via API
        :param launcher: Launcher object
        :param activity: Activity object
        :param missing: True if missing enrolment is expected or False otherwise
        :return: List of enrolments
    """
    # Authenticate with learner launcher credentials
    client = auth_utils.client_with_launcher_credentials(launcher)

    # Get the user profile
    profile = auth_utils.get_profile(client)
    assert "LEARNER" in profile['roles']

    # Get the list of enrolments
    institution_id = profile['institution']['id']
    list_enrolments_resp = client.get(
        '/api/v2/institution/{}/learner/{}/enrolment/?activity_id={}'.format(
            institution_id, profile['id'], activity['id']
        )
    )
    assert list_enrolments_resp.status_code == 200
    assert list_enrolments_resp.data['missing_enrolments'] == missing

    return list_enrolments_resp.data


def api_lapi_perform_enrolment(launcher, instruments):
    """
        The learner perform enrolment for missing instruments, sending data using LAPI and API
        :param launcher: Learner launcher object
        :param instruments: List of instruments to enrol
        :return: List of pending tasks assigned to providers
    """
    # List simulating the storage queue
    pending_tasks_storage = []
    # List simulating the validation queue
    pending_tasks_validation = []
    # Object simulating the providers validation queues
    pending_provider_tasks_validation = []

    def create_sample_test(*args, **kwargs):
        pending_tasks_storage.append(('create_sample', args, kwargs))

    def validate_request_test(*args, **kwargs):
        pending_tasks_validation.append(('validate_request', args, kwargs))

    def validate_request_prov_test(*args, **kwargs):
        pending_provider_tasks_validation.append(('validate_request', args, kwargs))

    # Authenticate with learner launcher credentials
    client = auth_utils.client_with_launcher_credentials(launcher)

    # Get the user profile
    profile = auth_utils.get_profile(client)
    assert "LEARNER" in profile['roles']

    # Get learner data (it is injected with the JS script)
    inst_id = profile['institution']['id']
    learner_data_resp = client.get('/api/v2/institution/{}/learner/{}/'.format(inst_id, profile['id']))
    assert learner_data_resp.status_code == 200
    learn_id = learner_data_resp.data['learner_id']

    # Create a data object simulating a sensor capture
    sensor_data = {
        'learner_id': learn_id,
        'instruments': [],
        'metadata': {
            'mimetype': 'some/mimetype',
            'filename': None,
            'created_at': None,
            'context': {}
        },
        'data': None
    }
    with mock.patch('tesla_ce.tasks.requests.enrolment.create_sample.apply_async', create_sample_test):
        for inst in instruments:
            sensor_data['instruments'] = [inst]
            # Send samples
            for _ in range(4*inst):
                sensor_data['data'] = get_random_string(50)
                sensor_data['created_at']: timezone.now()
                data_sent_resp = client.post(
                    '/lapi/v1/enrolment/{}/{}/'.format(inst_id, learn_id),
                    data=sensor_data,
                    format='json'
                )
                assert data_sent_resp.status_code == 200
                assert data_sent_resp.data['status'] == 'OK'
                if learn_id not in samples_list:
                    samples_list[learn_id] = []
                samples_list[learn_id].append(data_sent_resp.data['path'])

    # Run storage tasks
    with mock.patch('tesla_ce.tasks.requests.enrolment.validate_request.apply_async', validate_request_test):
        from tesla_ce.tasks.requests.enrolment import create_sample
        for task in pending_tasks_storage:
            assert task[0] == 'create_sample'
            create_sample(**task[2]['kwargs'])

    # Run validation tasks
    with mock.patch('tesla_ce.tasks.requests.enrolment.validate_request.apply_async', validate_request_prov_test):
        from tesla_ce.tasks.requests.enrolment import validate_request
        for task in pending_tasks_validation:
            assert task[0] == 'validate_request'
            validate_request(*task[1][0])

    # Return pending provider validation tasks
    return pending_provider_tasks_validation


def lapi_check_requests_status(launcher, expected_status):
    """
        The learner check the status of sent requests using LAPI
        :param launcher: Learner launcher object
        :param expected_status: The expected status for requests
    """
    lapi_check_sample_status(launcher, expected_status, requests_list)


def lapi_check_sample_status(launcher, expected_status, input_list=None):
    """
        The learner check the status of sent samples using LAPI
        :param launcher: Learner launcher object
        :param expected_status: The expected status for samples
        :param input_list: List of samples to check
    """
    if isinstance(expected_status, str):
        expected_status = [expected_status]

    if input_list is None:
        input_list = samples_list

    # Authenticate with learner launcher credentials
    client = auth_utils.client_with_launcher_credentials(launcher)

    # Get the user profile
    profile = auth_utils.get_profile(client)
    assert "LEARNER" in profile['roles']

    # Get learner data (it is injected with the JS script)
    inst_id = profile['institution']['id']
    learner_data_resp = client.get('/api/v2/institution/{}/learner/{}/'.format(inst_id, profile['id']))
    assert learner_data_resp.status_code == 200
    learn_id = learner_data_resp.data['learner_id']

    # Check that this learner have sent data
    assert learn_id in input_list
    status_resp = client.post('/lapi/v1/status/{}/{}/'.format(inst_id, learn_id),
                              data={
                                  'samples': input_list[learn_id],
                                  'learner_id': learn_id,
                              }, format='json')
    assert status_resp.status_code == 200
    assert len(input_list[learn_id]) == len(status_resp.data)
    for stat in status_resp.data:
        assert stat['status'] in expected_status


def get_data_object_from_session(assessment_session):
    """
        Get data object from assessment session
        :param assessment_session: Assessment session object
        :return: Data object
    """
    # Get the data object
    injected_data_resp = requests.get(assessment_session['data']['data'], verify=False)
    assert injected_data_resp.status_code == 200
    data = injected_data_resp.json()

    # Authenticate with learner launcher credentials
    assert 'launcher' in data
    client = auth_utils.client_with_launcher_credentials(data['launcher'])

    # Get the user profile
    profile = auth_utils.get_profile(client)
    assert "LEARNER" in profile['roles']

    assert data['learner']['id'] == profile['id']
    assert data['learner']['institution_id'] == profile['institution']['id']
    assert data['learner']['first_name'] == profile['first_name']
    assert data['learner']['last_name'] == profile['last_name']

    # Check that data contains the learner credentials
    assert 'token' in data
    assert 'access_token' in data['token']
    assert 'refresh_token' in data['token']

    # Check that data contains options
    assert 'options' in data
    assert 'floating_menu_initial_pos' in data['options']

    return injected_data_resp.json()


def lapi_learner_perform_activity(assessment_session):
    """
        The learner perform the activity, sending information from sensors using the LAPI
        :param assessment_session: Assessment session object
        :return: Tuple with the list of tasks pending to be executed by providers and performed activity
    """
    # Performed activity
    performed_activity = None

    # List simulating the storage queue
    pending_tasks_storage = []
    # List simulating the verification queue
    pending_tasks_verification = []
    # Object simulating the providers verification queues
    pending_provider_tasks_verification = []

    def create_request_test(*args, **kwargs):
        pending_tasks_storage.append(('create_request', args, kwargs))

    def verify_request_test(*args, **kwargs):
        pending_tasks_verification.append(('verify_request', args, kwargs))

    def verify_request_prov_test(*args, **kwargs):
        pending_provider_tasks_verification.append(('verify_request', args, kwargs))

    # Get the assessment session data
    session_data = get_data_object_from_session(assessment_session)

    # Authenticate with learner launcher credentials
    client = auth_utils.client_with_token_credentials(session_data['token']['access_token'],
                                                      session_data['token']['refresh_token'])

    # Get the required data
    institution_id = session_data['learner']['institution_id']
    learner_id = session_data['learner']['learner_id']
    sensors = session_data['sensors']

    # Create a data object simulating a sensor capture
    sensor_data = {
        'learner_id': learner_id,
        'course_id': session_data['activity']['course']['id'],
        'activity_id': session_data['activity']['id'],
        'session_id': session_data['session_id'],
        'instruments': [],
        'metadata': {
            'mimetype': 'some/mimetype',
            'filename': None,
            'created_at': None,
            'context': {}
        },
        'data': None
    }
    with mock.patch('tesla_ce.tasks.requests.verification.create_request.apply_async', create_request_test):
        for sensor in sensors:
            if sensor == 'activity':
                # Activities are sent by VLE when learner submit them. Just create a random activity
                performed_activity = {
                    'filename': get_random_string(50),
                    'content': get_random_string(2048),
                    'mimetype': 'activity/mimetype'
                }
            else:
                sensor_data['instruments'] = sensors[sensor]
                sensor_data['metadata']['mimetype'] = '{}/mimetype'.format(sensor)
                # Send samples
                for capture_id in range(10):
                    sensor_data['data'] = get_random_string(50)
                    sensor_data['metadata']['created_at'] = timezone.now()
                    sensor_data['metadata']['context']['sequence'] = capture_id
                    data_sent_resp = client.post(
                        '/lapi/v1/verification/{}/{}/'.format(institution_id, learner_id),
                        data=sensor_data,
                        format='json'
                    )
                    assert data_sent_resp.status_code == 200
                    assert data_sent_resp.data['status'] == 'OK'
                    if learner_id not in requests_list:
                        requests_list[learner_id] = []
                    requests_list[learner_id].append(data_sent_resp.data['path'])

            # Learner can refresh the token
            new_token = auth_utils.refresh_token(session_data['token']['access_token'],
                                                 session_data['token']['refresh_token'])
            client = auth_utils.client_with_token_credentials(new_token['access_token'],
                                                              new_token['refresh_token'])

    # Run storage tasks
    with mock.patch('tesla_ce.tasks.requests.verification.verify_request.apply_async', verify_request_test):
        from tesla_ce.tasks.requests.verification import create_request
        for task in pending_tasks_storage:
            assert task[0] == 'create_request'
            create_request(**task[2]['kwargs'])

    # Run verification tasks
    with mock.patch('tesla_ce.tasks.requests.verification.verify_request.apply_async', verify_request_prov_test):
        from tesla_ce.tasks.requests.verification import verify_request
        for task in pending_tasks_verification:
            assert task[0] == 'verify_request'
            verify_request(task[1][0][0])

    return pending_provider_tasks_verification, performed_activity



#  Copyright (c) 2021 Mireia Bellot
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Test module for institution course learners management """
import pytest

import tests.utils


@pytest.mark.django_db
def test_api_institution_course_learners(rest_api_client, user_global_admin, institution_course_test_case):
    # institution
    institution_id = institution_course_test_case['institution'].id
    # institution user
    institution_user = institution_course_test_case['user'].institutionuser
    institution_user_uid = institution_user.uid
    # instructor
    instructor = institution_course_test_case['instructor']
    instructor_user = instructor.institutionuser
    instructor_user_id = instructor_user.id
    # learner
    learner = institution_course_test_case['learner']
    learner_user = learner.institutionuser
    learner_user_id = learner_user.id

    pytest.skip('TODO')

    # Get the list of courses for a normal Institution User not belonging to any course
    institution_user.inst_admin = False
    institution_user.data_admin = False
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)
    courses_url = '/api/v2/institution/{}/course/'.format(institution_id)
    inst_admin_resp = tests.utils.get_rest_api_client(rest_api_client, courses_url,
                                                      'List Courses Inst User', 'RESPONSE:', 200)
    assert inst_admin_resp['count'] == 0

    # Get the list of courses for a learner user and select her course_id
    rest_api_client.force_authenticate(user=learner_user)
    str_module = 'List of Learner\'s Courses (Learner user privileges)'
    str_message_learner = 'RESPONSE institution_id:{} learner_user_id:{}'.format(institution_id, learner_user_id)
    courses_list = tests.utils.get_rest_api_client(rest_api_client, courses_url,
                                                   str_module, str_message_learner, 200)
    n_courses = courses_list['count']
    assert n_courses == 1
    # Select first course
    course_id = courses_list['results'][0]['id']

    # An Instructor user adds a learner to course.
    rest_api_client.force_authenticate(user=instructor_user)
    str_path_learner = '{}{}/learner/'.format(courses_url, course_id)
    str_data = {'uid': institution_user.uid}
    tests.utils.post_rest_api_client(rest_api_client, str_path_learner, str_data,
                                     'Add new instructor to a course', 'RESPONSE: ', 201)
    str_module = 'List learners from a course (Instructor user privileges)'
    str_message_instructor = 'RESPONSE institution_id:{} course_id:{} instructor_user_id:{}'.format(institution_id,
                                                                                                    course_id,
                                                                                                    instructor_user_id)
    body = tests.utils.get_rest_api_client(rest_api_client, str_path_learner,
                                           str_module, str_message_instructor, 200)
    n_learners = body['count']
    assert n_learners == 2

    # Learner cannot add/delete new learner to a course
    rest_api_client.force_authenticate(user=learner_user)
    str_data = {'uid': institution_user_uid}
    tests.utils.post_rest_api_client(rest_api_client, str_path_learner, str_data,
                                     'Add new learner to a course (Learner user privilege)',
                                     'RESPONSE: ', 403)

    # Learner can read only her own information: List learners from a course (learner user privileges)
    str_module = 'List learners from a course (Learner user privileges)'
    body = tests.utils.get_rest_api_client(rest_api_client, str_path_learner,
                                           str_module, str_message_learner, 200)
    assert body['count'] == 1
    assert body['results'][0]['id'] == learner_user_id

    # An instructor user adds an instructor to course.
    rest_api_client.force_authenticate(user=instructor_user)
    str_path_instructor = '{}{}/instructor/'.format(courses_url, course_id)
    str_data = {'uid': institution_user.uid}
    tests.utils.post_rest_api_client(rest_api_client, str_path_instructor, str_data,
                                     'Add new instructor to a course', 'RESPONSE: ', 201)
    str_module = 'List instructor from a course (Instructor user privileges)'
    body = tests.utils.get_rest_api_client(rest_api_client, str_path_instructor,
                                           str_module, str_message_instructor, 200)
    n_instructors = body['count']
    assert n_instructors == 2

    # Learner cannot add/delete new instructor to a course
    rest_api_client.force_authenticate(user=learner_user)
    tests.utils.post_rest_api_client(rest_api_client, str_path_instructor, str_data,
                                     'Add new learner to a course (Learner user privilege)',
                                     'RESPONSE: ', 403)

    # Learner can read only her own information: List instructors from a course (learner user privileges)
    str_module = 'List of Course\'s Instructors (Learner user privileges)'
    instructors_list = tests.utils.get_rest_api_client(rest_api_client, str_path_instructor,
                                                       str_module, str_message_learner, 200)
    assert n_instructors == instructors_list['count']

    # TODO Delete learner: remove from course but not from system (institution admin)
    # 666 TODO? remove second test learner from the course?



#!/usr/bin/env python

#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Django's command-line utility for administrative tasks."""
import os
import sys


def main():
    """
        Run DJango commands using Configurations library
    """
    os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'tesla_ce.settings')
    os.environ.setdefault('DJANGO_CONFIGURATION', 'Production')

    try:
        from configurations.management import execute_from_command_line

    except ImportError as exc:
        raise ImportError(
            "Couldn't import Django. Are you sure it's installed and "
            "available on your PYTHONPATH environment variable? Did you "
            "forget to activate a virtual environment?"
        ) from exc
    execute_from_command_line(sys.argv)


if __name__ == '__main__':
    main()



#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Report Activity Instrument model module."""
from enum import Enum
import json

from django.core.files.base import ContentFile
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.utils.translation import gettext_lazy as _

from .base_model import BaseModel
from .instrument import Instrument
from .report_activity import REPORT_ALERT_LEVEL
from .report_activity import ReportActivity
from .request_provider_result import RequestProviderResult
from .enrolment import Enrolment


class ResultFacts(Enum):
    """ Facts extracted from results """
    POSITIVE_LEARNER_RESULT_ABOVE_THRESHOLD = 1
    POSITIVE_LEARNER_RESULT_GOOD_INSTRUMENT = 2
    POSITIVE_LEARNER_RESULT_GOOD_ACTIVITY = 3
    POSITIVE_LEARNER_RESULT_FREQUENT = 4
    POSITIVE_CONFIDENCE_HIGH = 5
    NEUTRAL_MISSING_INFORMATION = 0
    NEGATIVE_LEARNER_RESULT_BELOW_THRESHOLD = -1
    NEGATIVE_LEARNER_RESULT_BAD_INSTRUMENT = -2
    NEGATIVE_LEARNER_RESULT_BAD_ACTIVITY = -3
    NEGATIVE_LEARNER_RESULT_NOT_FREQUENT = -4
    NEGATIVE_CONFIDENCE_LOW = -5


class ReportActivityInstrument(BaseModel):
    """ Report Activity model. """

    report = models.ForeignKey(ReportActivity, null=False, blank=False, on_delete=models.CASCADE,
                               help_text=_('Related Activity Report.'))

    instrument = models.ForeignKey(Instrument, null=False, blank=False, on_delete=models.CASCADE,
                                   help_text=_('Instrument related to this report detail.'))

    enrolment = models.SmallIntegerField(null=None, help_text=_('Enrolment percentage'))
    confidence = models.SmallIntegerField(null=None, help_text=_('Confidence percentage'))
    result = models.SmallIntegerField(null=None, help_text=_('Result percentage'))

    identity_level = models.SmallIntegerField(choices=REPORT_ALERT_LEVEL, null=False, default=0,
                                              help_text=_('Alert level for learner identity.'))
    content_level = models.SmallIntegerField(choices=REPORT_ALERT_LEVEL, null=False, default=0,
                                             help_text=_('Alert level for content authorship.'))
    integrity_level = models.SmallIntegerField(choices=REPORT_ALERT_LEVEL, null=False, default=0,
                                               help_text=_('Alert level for system integrity.'))

    audit_data = models.FileField(max_length=250, null=True, blank=False,
                                  help_text=_("Path to the audit data."))

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        unique_together = (('report', 'instrument'),)

    def __repr__(self):
        return "<ReportActivityInstrument(id='%r', report_id='%r' instrument_id='%r')>" % (
            self.id, self.report_id, self.instrument_id)

    def get_activity_histogram(self):
        """
            Get the activity histogram
            :return: Histogram
        """
        return list(self.instrument.histogramactivityinstrument_set.values_list(
            'b0', 'b1', 'b2', 'b3', 'b4', 'b5', 'b6', 'b7', 'b8', 'b9').first())

    def get_learner_histogram(self):
        """
            Get the learner histogram
            :return: Histogram
        """
        return list(self.report.learner.histogramlearnerinstrument_set.filter(
            instrument=self.instrument
        ).values_list('b0', 'b1', 'b2', 'b3', 'b4', 'b5', 'b6', 'b7', 'b8', 'b9').first())

    def get_instrument_polarity(self):
        """
            Return the instrument polarity
            :return: 1 for normal polarity and -1 if inverted polarity
        """
        return 2 * int(not self.instrument.provider_set.first().inverted_polarity) - 1

    def get_result_bean(self):
        """
            Get the histogram bin where current result is included
            :return: Bin number
        """
        return min(int(self.result/10), 9)

    def get_thresholds(self):
        """
            Get the reference thresholds for current instrument
            :return: Reference thresholds
        """
        prov_thr = self.instrument.provider_set.first()
        return {
            "warning_below": prov_thr.warning_below,
            "alert_below": prov_thr.alert_below
        }

    def get_prob_learner(self):
        """
            Compute the probability to have current value for the learner
            :return: Probability value
        """
        hist_bin = self.get_result_bean()
        hist = self.get_learner_histogram()
        acc = hist[hist_bin]
        if hist_bin > 0:
            acc += hist[hist_bin - 1] / 2.0
        if hist_bin < 9:
            acc += hist[hist_bin + 1] / 2.0

        return acc / sum(hist)

    def get_prob_context(self):
        """
            Compute the probability to have current value in the context
            :return: Probability value
        """
        hist_bin = self.get_result_bean()
        hist = self.get_activity_histogram()
        acc = hist[hist_bin]
        if hist_bin > 0:
            acc += hist[hist_bin - 1] / 2.0
        if hist_bin < 9:
            acc += hist[hist_bin + 1] / 2.0

        return acc / sum(hist)

    def get_h_prob_learner(self):
        """
            Compute the probability to have better value for the learner
            :return: Probability value
        """
        hist_bin = self.get_result_bean()
        hist = self.get_learner_histogram()
        if self.get_instrument_polarity() > 0:
            acc = sum(hist[hist_bin + 1: 10])
        else:
            acc = sum(hist[0:hist_bin])

        return acc / sum(hist)

    def get_h_prob_context(self):
        """
            Compute the probability to have better value in the context
            :return: Probability value
        """
        hist_bin = self.get_result_bean()
        hist = self.get_activity_histogram()
        if self.get_instrument_polarity() > 0:
            acc = sum(hist[hist_bin + 1: 10])
        else:
            acc = sum(hist[0:hist_bin])

        return acc / sum(hist)

    def get_facts(self):
        """
            Get facts from results
            :return: Dictionary with facts
        """
        if self.confidence < 0.0001 or self.result is None:
            # Not enough information
            facts = {
                'neutral': [ResultFacts.NEUTRAL_MISSING_INFORMATION.name]
            }
        else:
            facts = {
                'positive': [],
                'negative': []
            }
            # Check confidence
            if self.confidence < 0.75:
                facts['negative'].append(ResultFacts.NEGATIVE_CONFIDENCE_LOW.name)
            else:
                facts['positive'].append(ResultFacts.POSITIVE_CONFIDENCE_HIGH.name)
            # Check thresholds
            thresholds = self.get_thresholds()
            inst_polarity = self.get_instrument_polarity()
            polarized_result = self.result * inst_polarity
            if thresholds["warning_below"] is not None and \
                    polarized_result >= thresholds["warning_below"] * inst_polarity:
                facts['positive'].append(ResultFacts.POSITIVE_LEARNER_RESULT_ABOVE_THRESHOLD.name)
            elif thresholds["alert_below"] is not None and \
                    polarized_result < thresholds["alert_below"] * inst_polarity:
                facts['negative'].append(ResultFacts.NEGATIVE_LEARNER_RESULT_BELOW_THRESHOLD.name)
            # Check probabilities
            h_prob_inst = self.get_h_prob_learner()
            if h_prob_inst < 0.4:
                facts['positive'].append(ResultFacts.POSITIVE_LEARNER_RESULT_GOOD_INSTRUMENT.name)
            elif h_prob_inst > 0.75:
                facts['negative'].append(ResultFacts.NEGATIVE_LEARNER_RESULT_BAD_INSTRUMENT.name)
            h_prob_act = self.get_h_prob_context()
            if h_prob_act < 0.4:
                facts['positive'].append(ResultFacts.POSITIVE_LEARNER_RESULT_GOOD_ACTIVITY.name)
            elif h_prob_act > 0.5:
                facts['negative'].append(ResultFacts.NEGATIVE_LEARNER_RESULT_BAD_ACTIVITY.name)
            if self.get_prob_learner() > 0.8:
                facts['positive'].append(ResultFacts.POSITIVE_LEARNER_RESULT_FREQUENT.name)

        return facts

    def update_audit(self):
        """
            Update the audit data for this report
        """
        if self.instrument.id == 1:
            # Build audit data for Face Recognition
            audit = self._build_fr_audit()
        else:
            # Other instruments will have no audit data
            audit = None

        audit_path = '{}/results/{}/{}/{}/report_audit_{}.json'.format(
            self.report.activity.vle.institution_id,
            self.report.learner.learner_id,
            self.report.activity.course.id,
            self.report.activity.id,
            self.instrument.acronym
        )

        if audit is not None:
            # Save the audit data
            self.audit_data.save(audit_path,
                                 ContentFile(json.dumps(audit, cls=DjangoJSONEncoder).encode('utf-8')))
        else:
            self.audit_data = None

        self.save()

    def _build_fr_audit(self):
        """
            Build audit data for Face Recognition instrument
            :return: Audit data object
        """
        provider_results = RequestProviderResult.objects.filter(provider__instrument=self.instrument,
                                                                request__learner=self.report.learner,
                                                                request__activity=self.report.activity
                                                                ).all().order_by('request__created_at', 'provider_id')
        audit = {
            'providers': {},
            'requests': {},
            'enrolment_samples': {},
        }

        # Add results for each of the requests
        for result in provider_results:
            if result.provider.id not in audit['providers']:
                audit['providers'][result.provider.id] = {
                    'id': result.provider.id,
                    'acronym': result.provider.acronym,
                    'name': result.provider.name,
                    'enrolment_samples': []
                }
            if result.request.id not in audit['requests']:
                req_result = result.request.requestresult_set.first()
                req_result_value = None
                req_result_status = None
                req_result_error_message = None
                req_result_message_code = None
                req_result_code = None
                if req_result is not None:
                    req_result_value = req_result.result
                    req_result_status = req_result.status
                    req_result_error_message = req_result.error_message
                    req_result_code = req_result.code
                    if req_result.message_code is not None:
                        req_result_message_code = req_result.message_code.code

                audit['requests'][result.request.id] = {
                    'id': result.request.id,
                    'created_at': result.request.created_at,
                    'error_message': result.request.error_message,
                    'session': result.request.session.id,
                    'result': req_result_value,
                    'status': req_result_status,
                    'error_message_res': req_result_error_message,
                    'message_code': req_result_message_code,
                    'code': req_result_code,
                    'results': {},
                }
                if result.request.message_code is not None:
                    audit['requests'][result.request.id]['message_code'] = result.request.message_code

                try:
                    request_data = json.loads(result.request.data.read().decode())
                    audit['requests'][result.request.id]['data'] = request_data['data']
                    audit['requests'][result.request.id]['metadata'] = request_data['metadata']
                except Exception:
                    audit['requests'][result.request.id]['data'] = None
                    audit['requests'][result.request.id]['metadata'] = None

            # Read the audit for this request
            try:
                provider_audit = json.loads(result.audit.read().decode())
            except Exception:
                provider_audit = None

            audit['requests'][result.request.id]['results'][result.provider.id] = {
                'status': result.status,
                'result': result.result,
                'audit': provider_audit
            }

        # Add enrolment data
        for provider_id in audit['providers']:
            enrolment = Enrolment.objects.get(provider_id=provider_id, learner=self.report.learner)
            for sample in enrolment.model_samples.filter(status=1).all():
                if sample.id not in audit['enrolment_samples']:
                    try:
                        sample_data = json.loads(sample.data.read().decode())
                    except Exception:
                        sample_data = None
                    audit['enrolment_samples'][sample.id] = sample_data
                audit['providers'][provider_id]['enrolment_samples'].append(sample.id)
                audit['providers'][provider_id]['enrolment_percentage'] = enrolment.percentage

        return audit



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Learner api serialize module."""
from django.utils import timezone
from rest_framework import exceptions
from rest_framework import serializers

from tesla_ce.models import Course
from tesla_ce.models import InformedConsent
from tesla_ce.models import InstitutionUser
from tesla_ce.models import Learner
from tesla_ce.models import User


class VLECourseLearnerConsentSerializer(serializers.ModelSerializer):
    """Learner Informed Consent serialize model module."""

    status = serializers.CharField(read_only=True)

    class Meta:
        model = InformedConsent
        fields = ["version", "status"]


class VLECourseLearnerSerializer(serializers.ModelSerializer):
    """Learner serialize model module."""

    username = serializers.CharField(read_only=True)
    consent = VLECourseLearnerConsentSerializer(read_only=True, allow_null=True, default=None)
    consent_accepted = serializers.DateTimeField(read_only=True)
    consent_rejected = serializers.DateTimeField(read_only=True)
    learner_id = serializers.UUIDField(read_only=True)
    last_login = serializers.DateTimeField(read_only=True)
    send = serializers.JSONField(read_only=True)
    ic_status = serializers.CharField(read_only=True)
    enrolment_status = serializers.JSONField(read_only=True)

    first_name = serializers.CharField(required=False, default=None)
    last_name = serializers.CharField(required=False, default=None)
    email = serializers.EmailField(required=False, default=None)
    institution_id = serializers.HiddenField(default=None)

    class Meta:
        model = Learner
        exclude = ["password", "is_superuser", "is_staff", "is_active", "date_joined", "groups",
                   "user_permissions", "institution", "login_allowed"]

    @staticmethod
    def _create_institution_user(attrs):
        """
            Create an institution user for given attributes
            :param attrs: Attributes for the new institution user
            :type attrs: dict

            :return: Institution user
            :rtype: InstitutionUser
        """
        # Check if the user exists
        try:
            # Create using existing user
            user = User.objects.get(email=attrs['email'])
            inst_user = InstitutionUser(user_ptr=user, institution_id=attrs['institution_id'])
            inst_user.created_at = timezone.now()
            inst_user.updated_at = timezone.now()
            inst_user.save_base(raw=True)
            inst_user = InstitutionUser.objects.get(id=inst_user.id)
        except User.DoesNotExist:
            # Create a new user
            if attrs['first_name'] is None or attrs['last_name'] is None or attrs['email'] is None:
                raise exceptions.ValidationError(
                    'Learner does not exist and information to create it is not provided'
                )
            attrs['username'] = attrs['email']
            inst_user = InstitutionUser.objects.create(**attrs)
            inst_user.save()
            inst_user = InstitutionUser.objects.get(id=inst_user.id)
        return inst_user

    def validate(self, attrs):
        """
            Validate the given attributes
            :param attrs: Attributes parsed from request
            :type attrs: dict
            :return: Validated attributes
            :rtype: dict
        """
        # Get related objects
        try:
            course = Course.objects.get(id=self.context['view'].kwargs['parent_lookup_course_id'])
            institution = course.vle.institution
            attrs['institution_id'] = institution.id
        except Course.DoesNotExist:
            raise exceptions.NotFound('Course not found')

        try:
            learner = Learner.objects.get(institution=institution, uid=attrs['uid'])
            validated_data = super().validate(attrs)
        except Learner.DoesNotExist:
            # Check if institution allows to create new learners
            if institution.disable_vle_learner_creation:
                raise exceptions.NotFound('Learner not found')

            # Search in institution users
            try:
                inst_user = InstitutionUser.objects.get(institution_id=attrs['institution_id'], uid=attrs['uid'])
            except InstitutionUser.DoesNotExist:
                inst_user = None

            # If does not exist, try to create
            if inst_user is None:
                if institution.disable_vle_user_creation:
                    # User creation is not allowed
                    raise exceptions.NotFound('User not found')
                inst_user = self._create_institution_user(attrs)

            if inst_user is None:
                raise exceptions.ValidationError('Error creating new associated user')

            # Create the learner object
            learner = Learner(institutionuser_ptr=inst_user,
                              institution_id=attrs['institution_id'],
                              joined_at=timezone.now())
            learner.save_base(raw=True)
            learner = Learner.objects.get(id=inst_user.id)

            # Apply validators
            validated_data = super().validate(attrs)

        # Store the objects
        validated_data['learner'] = learner
        validated_data['course'] = course

        return validated_data

    def create(self, validated_data):
        """
            Create a new learner in the course
            :param validated_data: Validated data
            :return: New instance
        """
        learner = None
        if 'learner' in validated_data:
            learner = validated_data['learner']
        if learner is None:
            raise exceptions.ValidationError('Error creating learner')
        validated_data['course'].learners.add(learner)
        return learner


class VLECourseActivityLearnerSerializer(serializers.ModelSerializer):
    """Learner serialize model module."""

    class Meta:
        model = Learner
        fields = ['id', 'first_name', 'last_name', 'email', 'uid', 'institution_id', 'learner_id']



#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Api urls module """

from django.urls import include
from django.urls import path
from rest_framework_extensions.routers import ExtendedSimpleRouter

from . import views

# Create the base router
router = ExtendedSimpleRouter()

# VLE router
vle_router = router.register(r'vle', views.VLEViewSet, basename='vle')
vle_instrument_router = vle_router.register(r'instrument',
                                            views.VLEInstrumentViewSet,
                                            basename='vle-instrument',
                                            parents_query_lookups=['vle_id'])
vle_course_router = vle_router.register(r'course',
                                        views.VLECourseViewSet,
                                        basename='vle-course',
                                        parents_query_lookups=['vle_id'])
vle_course_activity_router = vle_course_router.register(r'activity',
                                                        views.VLECourseActivityViewSet,
                                                        basename='vle-course-activity',
                                                        parents_query_lookups=['vle_id', 'course_id'])

vle_course_activity_instrument_router = vle_course_activity_router.register(r'instrument',
                                                                            views.VLECourseActivityInstrumentViewSet,
                                                                            basename='vle-course-activity-instrument',
                                                                            parents_query_lookups=['vle_id',
                                                                                                   'course_id',
                                                                                                   'activity_id'])
vle_course_activity_learner_router = vle_course_activity_router.register(r'learner',
                                                                         views.VLECourseActivityLearnerViewSet,
                                                                         basename='vle-course-activity-learner',
                                                                         parents_query_lookups=['vle_id',
                                                                                                'course_id',
                                                                                                'activity_id'])
vle_course_activity_learner_request_router = vle_course_activity_learner_router.register(
    r'request',
    views.VLECourseActivityLearnerRequestViewSet,
    basename='vle-course-activity-learner-request',
    parents_query_lookups=['vle_id',
                           'course_id',
                           'activity_id',
                           'learner_id'
                           ])

vle_course_activity_report_router = vle_course_activity_router.register(
    r'report',
    views.VLECourseActivityReportViewSet,
    basename='vle-course-activity-report',
    parents_query_lookups=['activity__vle_id', 'activity__course_id', 'activity_id'])

vle_course_learner_router = vle_course_router.register(r'learner',
                                                       views.VLECourseLearnerViewSet,
                                                       basename='vle-course-learner',
                                                       parents_query_lookups=['vle_id', 'course_id'])
vle_course_instructor_router = vle_course_router.register(r'instructor',
                                                          views.VLECourseInstructorViewSet,
                                                          basename='vle-course-instructor',
                                                          parents_query_lookups=['vle_id', 'course_id'])

# Institution router
institution_router = router.register(r'institution', views.InstitutionViewSet, basename='institution')
# Institution => Informed consent routers
institution_ic_router = institution_router.register(r'ic',
                                                    views.InstitutionInformedConsentViewSet,
                                                    basename='institution-ic',
                                                    parents_query_lookups=['institution_id']
                                                    )
institution_ic_docs_router = institution_ic_router.register(r'document',
                                                            views.InstitutionInformedConsentDocumentViewSet,
                                                            basename='institution-ic-docs',
                                                            parents_query_lookups=['institution_id',
                                                                                   'informed_consent_id']
                                                    )

# Institution => Course Groups router
institution_course_group_router = institution_router.register(r'group',
                                                              views.InstitutionCourseGroupViewSet,
                                                              basename='institution-group',
                                                              parents_query_lookups=['institution_id']
                                                              )
institution_course_group_course_router = institution_course_group_router.register(
    r'course',
    views.InstitutionCourseGroupCourseViewSet,
    basename='institution-group-course',
    parents_query_lookups=['institution_id', 'id']
)
# Institution => SEND router
institution_send_router = institution_router.register(r'send',
                                                      views.InstitutionSENDCategoryViewSet,
                                                      basename='institution-send',
                                                      parents_query_lookups=['institution_id']
                                                      )

# Learner router
institution_learner_router = institution_router.register(r'learner',
                                                         views.InstitutionLearnerViewSet,
                                                         basename='institution-learner',
                                                         parents_query_lookups=['institution_id']
                                                         )

# Instructor router
institution_instructor_router = institution_router.register(r'instructor',
                                                            views.InstitutionInstructorViewSet,
                                                            basename='institution-instructor',
                                                            parents_query_lookups=['institution_id']
                                                            )

# Instrument router
institution_instrument_router = institution_router.register(r'instrument',
                                                            views.InstitutionInstrumentViewSet,
                                                            basename='institution-instrument',
                                                            parents_query_lookups=['institution_id']
                                                            )

# User Interface router
institution_ui_router = institution_router.register(r'ui',
                                                    views.InstitutionUIOptionViewSet,
                                                    basename='institution-ui',
                                                    parents_query_lookups=['institution_id']
                                                    )

# Course router
institution_course_router = institution_router.register(r'course',
                                                        views.InstitutionCourseViewSet,
                                                        basename='institution-course',
                                                        parents_query_lookups=['vle__institution_id']
                                                        )

institution_course_learner_router = institution_course_router.register(r'learner',
                                                                       views.InstitutionCourseLearnerViewSet,
                                                                       basename='institution-course-learner',
                                                                       parents_query_lookups=[
                                                                           'vle__institution_id',
                                                                           'id'
                                                                       ])

institution_course_instructor_router = institution_course_router.register(r'instructor',
                                                                          views.InstitutionCourseInstructorViewSet,
                                                                          basename='institution-course-instructor',
                                                                          parents_query_lookups=[
                                                                              'vle__institution_id',
                                                                              'id'
                                                                          ])

institution_course_activity_router = institution_course_router.register(r'activity',
                                                                        views.InstitutionCourseActivityViewSet,
                                                                        basename='institution-course-activity',
                                                                        parents_query_lookups=['vle__institution_id',
                                                                                               'course_id']
                                                                        )

institution_course_activity_instrument_router = institution_course_activity_router.register(
    r'instrument',
    views.InstitutionCourseActivityInstrumentViewSet,
    basename='institution-course-activity-instrument',
    parents_query_lookups=['activity__vle__institution_id',
                           'activity__course_id',
                           'activity_id']
)

institution_course_activity_report_router = institution_course_activity_router.register(
    r'report',
    views.InstitutionCourseActivityReportViewSet,
    basename='institution-course-activity-report',
    parents_query_lookups=['activity__vle__institution_id',
                           'activity__course_id',
                           'activity_id']
)

institution_course_activity_report_audit_router = institution_course_activity_report_router.register(
    r'audit',
    views.InstitutionCourseActivityReportAuditViewSet,
    basename='institution-course-activity-report-audit',
    parents_query_lookups=['report__activity__vle__institution_id',
                           'report__activity__course_id',
                           'report__activity_id',
                           'report_id']
)

institution_course_activity_report_requests_router = institution_course_activity_report_router.register(
    r'request',
    views.InstitutionCourseActivityReportRequestViewSet,
    basename='institution-course-activity-report-request',
    parents_query_lookups=['activity__vle__institution_id',
                           'activity__course_id',
                           'activity_id',
                           'id']
)

# VLE router
institution_vle = institution_router.register(r'vle',
                                              views.InstitutionVLEViewSet,
                                              basename='institution-vle',
                                              parents_query_lookups=['institution_id']
                                              )

# User router
institution_user = institution_router.register(r'user',
                                               views.InstitutionUserViewSet,
                                               basename='institution-user',
                                               parents_query_lookups=['institution_id']
                                               )

# Learner => SEND router
institution_learner_send_router = institution_learner_router.register(r'send',
                                                                      views.InstitutionSENDLearnerViewSet,
                                                                      basename='institution-learner-send',
                                                                      parents_query_lookups=['learner__institution_id',
                                                                                             'learner_id']
                                                                      )

# Administration API
# Instrument router
admin_user_router = router.register(r'admin/user', views.AdminUserViewSet, basename='admin-user')

# Instrument router
admin_instrument_router = router.register(r'admin/instrument', views.InstrumentViewSet, basename='admin-instrument')

# Instrument => Provider
instrument_provider_router = admin_instrument_router.register(r'provider',
                                                              views.AdminProviderViewSet,
                                                              basename='admin-instrument-provider',
                                                              parents_query_lookups=['instrument_id']
                                                              )
# Institution router
admin_institution_router = router.register(r'admin/institution', views.InstitutionAdminViewSet,
                                           basename='admin-institution')

# UI Option router
admin_ui_router = router.register(r'admin/ui', views.UIOptionViewSet,
                                  basename='admin-ui')


# Providers API
provider_router = router.register(r'provider', views.ProviderViewSet, basename='provider')

# Provider => Enrolment
provider_enrolment_router = provider_router.register(r'enrolment',
                                                     views.ProviderEnrolmentViewSet,
                                                     basename='provider-enrolment',
                                                     parents_query_lookups=['provider_id']
                                                     )
# Provider => Enrolment => Sample
provider_enrolment_sample_router = provider_enrolment_router.register(r'sample',
                                                                      views.ProviderEnrolmentSampleViewSet,
                                                                      basename='provider-enrolment-sample',
                                                                      parents_query_lookups=[
                                                                          'instruments__provider__id',
                                                                          'learner__learner_id'
                                                                      ])

# Provider => Enrolment => Sample => Validation
provider_enrolment_sample_validation_router = provider_enrolment_sample_router.register(
    r'validation',
    views.ProviderEnrolmentSampleValidationViewSet,
    basename='provider-enrolment-sample-validation',
    parents_query_lookups=['provider_id', 'sample__learner__learner_id', 'sample_id']
)

# Provider => Verification Request
provider_verification_router = provider_router.register(r'request',
                                                     views.ProviderVerificationRequestResultViewSet,
                                                     basename='provider-request',
                                                     parents_query_lookups=['provider_id']
                                                     )

# Provider => Notification
provider_notification_router = provider_router.register(r'notification',
                                                        views.ProviderNotificationViewSet,
                                                        basename='provider-notification',
                                                        parents_query_lookups=['provider_id']
                                                        )

app_name = 'tesla_ce.apps.api'
urlpatterns = [
    path('auth/', include('tesla_ce.apps.api.auth.urls')),
    path('', include(router.urls)),
]



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Database Manager Module"""
import json
import os
import uuid

from django.db import connections
from django.db.migrations.executor import MigrationExecutor

from tesla_ce.lib import ConfigManager
from tesla_ce.lib.checks import check_database_connection
from tesla_ce.lib.exception import TeslaDatabaseException


class DatabaseManager:
    """ Manager class for Databases """

    def __init__(self, config=None):
        """
            Default constructor
            :param config: Configuration manager instance
            :type config: ConfigManager
        """
        if config is None:
            config = ConfigManager()

        # Store the configuration object
        self._config = config

    def initialize(self):
        """
            Initialize the database
        """
        # Get the connection status
        connection_status = self.check_connection()

        # Check database
        if not connection_status['default']['connected'] and not connection_status['admin']['connected']:
            raise TeslaDatabaseException('Cannot connect to the database {}@{}:{}/{}'.format(
                self._config.config.get('DB_USER'),
                self._config.config.get('DB_HOST'),
                self._config.config.get('DB_PORT'),
                self._config.config.get('DB_NAME')
            ))

        # Create TeSLA database if it does not exist
        if not connection_status['default']['connected']:
            # Create a password if it does not exist
            if self._config.config.get('DB_PASSWORD') is None:
                new_password = uuid.uuid4().__str__()
                self._config.config.set('DB_PASSWORD', new_password)
                self._config.save_configuration()
            # Create the database
            self.create_database()
            # Create the user
            self.create_user()

            # Check connection
            connection_status = self.check_connection()
            if not connection_status['default']['connected']:
                raise TeslaDatabaseException('Database and user creation failed')

            # Save created credentials
            self._config.save_configuration()

        # Apply migrations
        self.apply_migrations()

        # Update messages
        self.update_default_messages()

        # Set default institution
        self.update_default_institution()

        # Set default instruments
        self.update_default_instruments()

        # Create periodic tasks
        self.update_periodic_tasks()

    def check_connection(self):
        """
            Check database connection and access
        """
        return check_database_connection(config=self._config.config)

    def version(self):
        """
            Get the version of the database structure

            :return: Version of the database or None if no migration is applied
            :rtype: int
        """
        status = self._check_migrations()
        return status['current_version']

    def apply_migrations(self):
        """
            Apply pending migrations to the database

            :return: Result of the migration
        """
        status = self._check_migrations(admin=True)
        migration = status['executor'].migrate(targets=status['targets'],
                                               plan=status['plan'])
        return migration

    def create_database(self, name=None):
        """
            Create new database
        """
        engine = self._config.config.get('DB_ENGINE')
        creation_db_params = connections['admin'].get_connection_params()
        creation_db_params['db'] = ''
        creation_db = connections['admin'].get_new_connection(creation_db_params)
        cursor = creation_db.cursor()
        if name is None:
            name = self._config.config.get('DB_NAME')
        if engine == 'mysql':
            cursor.execute('CREATE DATABASE IF NOT EXISTS {};'.format(name))
        elif engine == 'postgresql':
            cursor.execute(
                'SELECT \'CREATE DATABASE {}\' WHERE NOT EXISTS (SELECT FROM pg_database WHERE datname = \'{}\')\\gexec'.format(
                    name, name)
            )
        else:
            raise TeslaDatabaseException('Unsuported engine {}'.format(engine))

    def create_user(self, db_name=None, db_user=None, db_password=None):
        """
            Create a new user for a database
        """
        # TODO: Implement different versions depending on Engine
        engine = self._config.config.get('DB_ENGINE')
        cursor = connections['admin'].cursor()

        if db_name is None:
            db_name = self._config.config.get('DB_NAME')
        if db_user is None:
            db_user = self._config.config.get('DB_USER')
        if db_password is None:
            db_password = self._config.config.get('DB_PASSWORD')

        if engine == 'mysql':
            cursor.execute('GRANT ALL PRIVILEGES ON {}.* TO \'{}\'@\'%\' IDENTIFIED BY \'{}\';'.format(
                db_name,
                db_user,
                db_password
            ))
        elif engine == 'postgresql':
            raise NotImplementedError('PostgreSQL user creation not implemented')
        else:
            raise TeslaDatabaseException('Unsuported engine {}'.format(engine))

    def get_status(self):
        """
            Get the database status

            :return: Object with status information
            :rtype: dict
        """
        return {
            'status': 0,
            'warnings': 0,
            'errors': 0,
            'info': {}
        }

    def _get_executor(self, admin=False):
        """
            Create a migration executor
            :return: Executor object ready to access default database
        """
        connection = connections['default']
        if admin:
            connection = connections['admin']
        executor = MigrationExecutor(connection)
        executor.loader.check_consistent_history(connection)
        return executor

    def _check_migrations(self, admin=False):
        """
            Check migrations status
            :return: Migration status report
            :rtype: dict
        """
        # Get the executor
        executor = self._get_executor(admin)

        # Check for conflicts
        conflicts = executor.loader.detect_conflicts()

        # Get target migrations for all applications
        targets = executor.loader.graph.leaf_nodes()

        # Compute a migration plan
        plan = executor.migration_plan(targets)

        # Get TeSLA CE migrations newest version
        new_version = None
        for target_desc in targets:
            if target_desc[0] == 'tesla_ce':
                new_version = int(target_desc[1].split('_')[0])

        # Get TeSLA CE migrations current version
        current_version = None
        for plan_desc in plan:
            if plan_desc[0].app_label == 'tesla_ce':
                version = int(plan_desc[0].name.split('_')[0])
                if current_version is None or version < current_version:
                    current_version = version
        if current_version is None:
            # There are no migrations to apply. We are on the target migration
            current_version = new_version
        else:
            # If the lower migration to apply is the first one, there is no migration applied
            if current_version == 1:
                current_version = None
            else:
                current_version = current_version - 1

        return {
            'executor': executor,
            'targets': targets,
            'plan': plan,
            'conflicts': conflicts,
            'current_version': current_version,
            'new_version': new_version
        }

    def update_default_institution(self):
        """
            Update the information for the default institution
        """
        from tesla_ce.models import Institution
        try:
            default_institution = Institution.objects.get(pk=1)
            default_institution.name = self._config.config.get('TESLA_INSTITUTION_NAME')
            default_institution.acronym = self._config.config.get('TESLA_INSTITUTION_ACRONYM')
            default_institution.save()
        except Institution.DoesNotExist:
            Institution.objects.create(name=self._config.config.get('TESLA_INSTITUTION_NAME'),
                                       acronym=self._config.config.get('TESLA_INSTITUTION_ACRONYM'))

    def update_default_instruments(self):
        """
            Update the information for the default instruments
        """
        from tesla_ce.models import Instrument
        fr_vr_schema = json.dumps(
                    {
                        'type': 'object',
                        'properties': {
                            'online': {
                                'type': 'boolean',
                                'title': 'Analyze learner identity during the assessment',
                                'default': True
                            },
                            'offline': {
                                'type': 'boolean',
                                'title': 'Analyze learner identity on the delivered assessment',
                                'default': False
                            }
                        }
                    }
                )
        default_instruments = (
            {
                'id': 1, 'name': 'Face Recognition', 'acronym': 'fr', 'queue': 'tesla_fr',
                'requires_enrolment': True, 'identity': True, 'originality': False, 'authorship': False,
                'description': 'Verify learner identity by means of face attributes.',
                'options_schema': fr_vr_schema
            },
            {
                'id': 2, 'name': 'Keystroke Dynamics Recognition', 'acronym': 'ks', 'queue': 'tesla_ks',
                'requires_enrolment': True, 'identity': True, 'originality': False, 'authorship': False,
                'description': 'Verify learner identity by means of keystroke patterns.',
                'options_schema': None
            },
            {
                'id': 3, 'name': 'Voice Recognition', 'acronym': 'vr', 'queue': 'tesla_vr',
                'requires_enrolment': True, 'identity': True, 'originality': False, 'authorship': False,
                'description': 'Verify learner identity by means of voice attributes.',
                'options_schema': fr_vr_schema
            },
            {
                'id': 4, 'name': 'Forensic Analysis', 'acronym': 'fa', 'queue': 'tesla_fa',
                'requires_enrolment': True, 'identity': True, 'originality': False, 'authorship': True,
                'description': 'Verify learner identity by means of writing style patterns.',
                'options_schema': None
            },
            {
                'id': 5, 'name': 'Plagiarism Detection', 'acronym': 'plag', 'queue': 'tesla_plag',
                'requires_enrolment': False, 'identity': False, 'originality': True, 'authorship': False,
                'description': 'Verify the originality of an assessment.',
                'options_schema': None
            }
        )
        for inst in default_instruments:
            try:
                inst_obj = Instrument.objects.get(pk=inst['id'])
                inst_obj.name = inst['name']
                inst_obj.description = inst['description']
                inst_obj.acronym = inst['acronym']
                inst_obj.queue = inst['queue']
                inst_obj.requires_enrolment = inst['requires_enrolment']
                inst_obj.identity = inst['identity']
                inst_obj.originality = inst['originality']
                inst_obj.authorship = inst['authorship']
                inst_obj.options_schema = inst['options_schema']
                inst_obj.save()
            except Instrument.DoesNotExist:
                Instrument.objects.create(**inst)

    def update_default_messages(self):
        """
            Update the default messages
        """
        from tesla_ce.models import Message, MessageLocale

        messages_file = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'data', 'messages.csv'))

        with open(messages_file, 'r', encoding='latin-1') as msg_file:
            headers = msg_file.readline().replace('\n', '')
            header_parts = headers.split(';')
            languages = header_parts[4:]
            line = msg_file.readline().replace('\n', '')
            while line is not None and len(line) > 0:
                line_parts = line.split(';')
                msg, created = Message.objects.update_or_create(code=line_parts[0],
                                                                defaults={
                                                                    'datatype': line_parts[1],
                                                                    'description': line_parts[2],
                                                                    'meaning': line_parts[3]
                                                                })
                translations = line_parts[4:]
                for lang_code, translation in zip(languages, translations):
                    if len(translation) > 0:
                        MessageLocale.objects.update_or_create(code=msg,
                                                               locale=lang_code,
                                                               defaults={'message': translation})
                line = msg_file.readline()

    def update_periodic_tasks(self):
        """
            Create the periodic tasks launched by the TeSLA Worker Beat
        """
        from django_celery_beat.models import IntervalSchedule, PeriodicTask

        # Get task intervals
        providers_notification = 30

        # Send notifications to providers
        task_name = 'tesla_ce.tasks.notification.providers.send_provider_notifications'
        queue = self._config.config.get('CELERY_QUEUE_DEFAULT')
        try:
            interval = IntervalSchedule.objects.get(every=providers_notification, period=IntervalSchedule.SECONDS)
        except IntervalSchedule.DoesNotExist:
            interval = IntervalSchedule.objects.create(every=providers_notification, period=IntervalSchedule.SECONDS)
        PeriodicTask.objects.update_or_create(name=task_name, task=task_name, defaults={
            'enabled': 1,
            'interval': interval,
            'description': 'Send notifications to providers',
            'queue': queue
        })




# Generated by Django 3.2.4 on 2021-06-16 17:59

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('tesla_ce', '0013_webhookclient_webhookmessage'),
    ]

    operations = [
        migrations.CreateModel(
            name='UIOption',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('route', models.CharField(db_index=True, help_text='Affected Route.', max_length=250)),
                ('enabled', models.BooleanField(blank=None, default=False, help_text='Status', null=None)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('institution', models.ForeignKey(default=None, help_text='Affected institution', null=True, on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.institution')),
                ('user', models.ForeignKey(default=None, help_text='Affected user', null=True, on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.institutionuser')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



#  Copyright (c) 2021 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Test module for institution course management """
import pytest

import tests.utils


@pytest.mark.django_db
def test_api_institution_course_list(rest_api_client, user_global_admin, institution_course_test_case):
    # Get general parameters
    institution_id = institution_course_test_case['institution'].id
    courses_url = '/api/v2/institution/' + str(institution_id) + '/course/'

    # Get the list of courses for the global admin
    user_global_admin.is_staff = True
    user_global_admin.save()
    rest_api_client.force_authenticate(user=user_global_admin)
    global_admin_resp = tests.utils.get_rest_api_client(
        rest_api_client, courses_url, 'List Courses Global Admin', 'RESPONSE:', 200)
    assert global_admin_resp['count'] == 1

    # Get the list of courses for an Institution Admin
    institution_user = institution_course_test_case['user'].institutionuser
    institution_user.inst_admin = True
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)
    inst_admin_resp = tests.utils.get_rest_api_client(rest_api_client, courses_url,
                                                      'List Courses Inst Admin', 'RESPONSE:', 200)
    assert inst_admin_resp['count'] == 1

    # Get the list of courses for a normal Institution User not belonging to any course
    institution_user = institution_course_test_case['user'].institutionuser
    institution_user.inst_admin = False
    institution_user.data_admin = False
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)
    inst_admin_resp = tests.utils.get_rest_api_client(rest_api_client, courses_url,
                                                      'List Courses Inst User', 'RESPONSE:', 200)
    assert inst_admin_resp['count'] == 0

    # Get the list of courses for a learner
    rest_api_client.force_authenticate(user=institution_course_test_case['learner'])
    learner_resp = tests.utils.get_rest_api_client(rest_api_client, courses_url,
                                                   'List Courses Learner', 'RESPONSE:', 200)
    assert learner_resp['count'] == 1

    # Get the list of courses for an instructor
    rest_api_client.force_authenticate(user=institution_course_test_case['instructor'])
    instructor_resp = tests.utils.get_rest_api_client(rest_api_client, courses_url,
                                                      'List Courses Instructor', 'RESPONSE:', 200)
    assert instructor_resp['count'] == 1

    # Get the list of courses for an unauthenticated user
    rest_api_client.force_authenticate(user=None)
    tests.utils.get_rest_api_client(rest_api_client, courses_url,
                                    'List Courses Unauthenticated', 'RESPONSE:', 403)


@pytest.mark.django_db
def test_api_institution_course_detail(rest_api_client, user_global_admin, institution_course_test_case):
    # Get general parameters
    courses_url = '/api/v2/institution/{}/course/{}/'.format(
        institution_course_test_case['institution'].id,
        institution_course_test_case['course'].id
    )

    # Get the list of courses for the global admin
    user_global_admin.is_staff = True
    user_global_admin.save()
    rest_api_client.force_authenticate(user=user_global_admin)
    global_admin_resp = tests.utils.get_rest_api_client(
        rest_api_client, courses_url, 'Course Info Global Admin', 'RESPONSE:', 200)
    assert global_admin_resp['id'] == institution_course_test_case['course'].id

    # Get the list of courses for an Institution Admin
    institution_user = institution_course_test_case['user'].institutionuser
    institution_user.inst_admin = True
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)
    inst_admin_resp = tests.utils.get_rest_api_client(rest_api_client, courses_url,
                                                      'Course Info Inst Admin', 'RESPONSE:', 200)
    assert inst_admin_resp['id'] == institution_course_test_case['course'].id

    # Get the list of courses for a normal Institution User not belonging to any course
    institution_user = institution_course_test_case['user'].institutionuser
    institution_user.inst_admin = False
    institution_user.data_admin = False
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)
    tests.utils.get_rest_api_client(rest_api_client, courses_url,
                                    'Course Info Inst User', 'RESPONSE:', 404)

    # Get the list of courses for a learner
    rest_api_client.force_authenticate(user=institution_course_test_case['learner'])
    learner_resp = tests.utils.get_rest_api_client(rest_api_client, courses_url,
                                                   'Course Info Learner', 'RESPONSE:', 200)
    assert learner_resp['id'] == institution_course_test_case['course'].id

    # Get the list of courses for an instructor
    rest_api_client.force_authenticate(user=institution_course_test_case['instructor'])
    instructor_resp = tests.utils.get_rest_api_client(rest_api_client, courses_url,
                                                      'Course Info Instructor', 'RESPONSE:', 200)
    assert instructor_resp['id'] == institution_course_test_case['course'].id

    # Get the list of courses for an unauthenticated user
    rest_api_client.force_authenticate(user=None)
    tests.utils.get_rest_api_client(rest_api_client, courses_url,
                                    'List Courses Unauthenticated', 'RESPONSE:', 403)



# Generated by Django 3.2.4 on 2021-07-15 19:50

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('tesla_ce', '0016_auto_20210709_1023'),
    ]

    operations = [
        migrations.AddField(
            model_name='reportactivity',
            name='data',
            field=models.FileField(help_text='Path to the content of this report.', max_length=250, null=True, upload_to=''),
        ),
        migrations.CreateModel(
            name='ReportActivitySession',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('identity_level', models.SmallIntegerField(choices=[(0, 'Pending'), (1, 'No Information'), (2, 'Ok'), (3, 'Warning'), (4, 'Alert')], default=0, help_text='Alert level for learner identity.')),
                ('content_level', models.SmallIntegerField(choices=[(0, 'Pending'), (1, 'No Information'), (2, 'Ok'), (3, 'Warning'), (4, 'Alert')], default=0, help_text='Alert level for content authorship.')),
                ('integrity_level', models.SmallIntegerField(choices=[(0, 'Pending'), (1, 'No Information'), (2, 'Ok'), (3, 'Warning'), (4, 'Alert')], default=0, help_text='Alert level for system integrity.')),
                ('total_requests', models.IntegerField(default=0, help_text='Number of requests on this session.')),
                ('pending_requests', models.IntegerField(default=0, help_text='Number of pending requests on this session.')),
                ('valid_requests', models.IntegerField(default=0, help_text='Number of valid requests on this session.')),
                ('processed_requests', models.IntegerField(default=0, help_text='Number of processed requests on this session.')),
                ('data', models.FileField(default=None, help_text='Path to the content of this report.', max_length=250, null=True, upload_to='')),
                ('closed_at', models.DateTimeField(default=None, help_text='When the session report was closed', null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('report', models.ForeignKey(help_text='Related Activity Report.', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.reportactivity')),
                ('session', models.ForeignKey(help_text='Assessment Session related to this report.', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.assessmentsession', unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]
